"""Helper module for launching the server."""
import logging
import os
import subprocess
import sys
from typing import Optional

import ansys.meshing.prime.internals.config as config
import ansys.meshing.prime.internals.defaults as defaults
import ansys.meshing.prime.internals.utils as utils
from ansys.meshing.prime.internals.client import Client

try:
    import ansys.platform.instancemanagement as pypim
    from simple_upload_server.client import Client as FileClient

    config.set_has_pim(pypim.is_configured())
except ModuleNotFoundError:
    pass

__all__ = ['launch_prime', 'launch_server_process']


def get_install_locations():
    supported_versions = ['241']
    awp_roots = {ver: os.environ.get(f'AWP_ROOT{ver}', '') for ver in supported_versions}
    installed_versions = {
        ver: os.path.join(path, 'meshing', 'Prime')
        for ver, path in awp_roots.items()
        if path and os.path.isdir(path)
    }
    installed_versions = {
        ver: path for ver, path in installed_versions.items() if path and os.path.isdir(path)
    }
    if installed_versions:
        return installed_versions


def get_ansys_prime_server_root():
    available_versions = get_install_locations()
    if not available_versions:
        return None

    version = max(available_versions.keys())
    prime_root = available_versions[version]
    return prime_root


def launch_server_process(
    prime_root: Optional[str] = None,
    ip: str = defaults.ip(),
    port: int = defaults.port(),
    n_procs: Optional[int] = None,
    **kw,
) -> subprocess.Popen:
    """Launch a server process for Ansys Prime Server.

    Parameters
    ----------
    prime_root : str, optional
        Root directory for Ansys Prime Server.
    ip: str
        IP address to start the server at. The default is ``127.0.0.1``.
    port: int
        Port at which the server is started. The default is ``50055``.
    n_procs: Optional[int]
        When running in distributed mode, the number of distributed
        processes to spawn. The default is ``None``, in which case
        the server is launched as the only process (normal mode). The
        process marked as ``Node 0`` hosts the gRPC server.

    Returns
    -------
    subprocess.Popen
        Instance of the subprocess that is launched.

    Raises
    ------
    FileNotFoundError
        When there is an error in the file paths used to launch the server.
    """
    if prime_root is None:
        prime_root = get_ansys_prime_server_root()
        if prime_root is None:
            raise FileNotFoundError('No valid Ansys Prime Server found to launch.')
    else:  # verify if the file exists
        if not os.path.isdir(prime_root):
            raise FileNotFoundError('Invalid exec_file path.')

    script_ext = 'bat' if os.name == 'nt' else 'sh'
    run_prime_script = f'runPrime.{script_ext}'

    exec_path = os.path.join(prime_root, run_prime_script)
    if not os.path.isfile(exec_path):
        raise FileNotFoundError(f'{run_prime_script} not found in {prime_root}')

    server_args = [exec_path, 'server']

    if kw is None:
        kw = {}

    enable_python_server = kw.get('server', 'release')
    if not isinstance(enable_python_server, str):
        raise ValueError(
            'Hidden option to run python server needs to be a string.'
            'Potential options are: release (default), debug and python'
        )

    enable_python_server = enable_python_server.lower()
    if enable_python_server not in ('python', 'release', 'debug'):
        raise ValueError('server option needs to be one of release, debug or python.')

    if enable_python_server == 'python':
        script_dir = os.path.join(prime_root, 'scripts')
        server_path = os.path.join(script_dir, 'PrimeGRPC.py')
        if not os.path.isfile(server_path):
            raise FileNotFoundError(f'PrimeGRPC.py not found in {script_dir}')
        server_args.append('-python')

    if enable_python_server == 'debug':
        server_args.append('-debug')

    server_args.append(f'--ip={ip}')
    server_args.append(f'--port={port}')
    if n_procs is not None and isinstance(n_procs, int):
        server_args.append(f'-np')
        server_args.append(f'{n_procs}')

    kwargs = {
        'stdin': subprocess.DEVNULL,
    }
    if "PRIME_ENABLE_VERBOSITY" not in os.environ:
        kwargs['stdout'] = subprocess.DEVNULL
        kwargs['stderr'] = subprocess.DEVNULL
    if sys.platform.startswith('win32'):
        kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP

    logging.getLogger('PyPrimeMesh').info('Launching Ansys Prime Server')
    server = subprocess.Popen(server_args, **kwargs)
    return server


def launch_remote_prime(
    version: Optional[str] = None, timeout: float = defaults.connection_timeout()
):
    """Create a remote instance of Ansys Prime Server using PyPIM.

    This method is used if Ansys Lab creates a remote instance of Ansys Prime Server.
    This method creates a file transfer service that is available on Ansys Lab.
    """
    if version is None:
        version = 'latest'

    pim = pypim.connect()
    instance = pim.create_instance(product_name='prime', product_version=version)
    instance.wait_for_ready()

    channel = instance.build_grpc_channel(
        options=[
            ('grpc.max_send_message_length', defaults.max_message_length()),
            ('grpc.max_receive_message_length', defaults.max_message_length()),
        ]
    )

    client = Client(channel=channel, timeout=timeout)
    file_service = FileClient(
        token='token',
        url=instance.services['http-simple-upload-server'].uri,
        headers=instance.services['http-simple-upload-server'].headers,
    )

    client.pim_client = pim
    client.remote_instance = instance
    model = client.model
    model._unfreeze()
    setattr(model, 'file_service', file_service)
    model._freeze()

    return client


def launch_prime(
    prime_root: Optional[str] = None,
    ip: str = defaults.ip(),
    port: int = defaults.port(),
    timeout: float = defaults.connection_timeout(),
    n_procs: Optional[int] = None,
    version: Optional[str] = None,
    **kwargs,
):
    """Launch an instance of Ansys Prime Server and get a client for it.

    Parameters
    ----------
    prime_root: Optional[str]
        Root directory for Ansys Prime Server.
    ip: str
        IP address to start the server at. The default is ``127.0.0.1``.
    port: int
        Port at which the server is started. The default is ``50055``.
    timeout: float
        Maximum time in seconds to wait for the client to connect to the server.
        The default is ``20.0``.
    n_procs: Optional[int]
        When running in distributed mode, the number of distributed
        processes to spawn. The default is ``None``, in which case
        the server is launched as the only process (normal mode). The
        process marked as ``Node 0`` hosts the gRPC server.

    Returns
    -------
    Client
        Instance of the client that is connected to the launched server.

    Raises
    ------
    FileNotFoundError
        When there is an error in file paths used to launch the server.
    ConnectionError
        When there is an error in connecting to the gRPC server.
    """
    if config.has_pim():
        return launch_remote_prime(version=version, timeout=timeout)

    # Check for port availability on local host
    if ip == defaults.ip():
        port = utils.get_available_local_port(port)

    launch_container = bool(int(os.environ.get('PYPRIMEMESH_LAUNCH_CONTAINER', '0')))
    if launch_container:
        container_name = utils.make_unique_container_name('ansys-prime-server')
        utils.launch_prime_github_container(port=port, name=container_name, version=version)
        config.set_using_container(True)
        client = Client(port=port, timeout=timeout)
        client.container_name = container_name
        return client

    server = launch_server_process(
        prime_root=prime_root, ip=ip, port=port, n_procs=n_procs, **kwargs
    )

    return Client(server_process=server, ip=ip, port=port, timeout=timeout)



""" Auto-generated file. DO NOT MODIFY """
import enum
from typing import Dict, Any, Union, List, Iterable
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.internals import utils
from ansys.meshing.prime.autogen.coreobject import *
import numpy as np

from ansys.meshing.prime.params.primestructs import *

class MeshStackerResults(CoreObject):
    """Results associated with the mesh stacker operations.
    """
    _default_params = {}

    def __initialize(
            self,
            error_code: ErrorCode,
            error_codes_per_topo_volume: Iterable[int],
            non_stackable_faces: Iterable[int],
            non_stackable_edges: Iterable[int],
            small_features: Iterable[int],
            base_face_ids: Iterable[int],
            size_control_ids: Iterable[int]):
        self._error_code = ErrorCode(error_code)
        self._error_codes_per_topo_volume = error_codes_per_topo_volume if isinstance(error_codes_per_topo_volume, np.ndarray) else np.array(error_codes_per_topo_volume, dtype=np.int32) if error_codes_per_topo_volume is not None else None
        self._non_stackable_faces = non_stackable_faces if isinstance(non_stackable_faces, np.ndarray) else np.array(non_stackable_faces, dtype=np.int32) if non_stackable_faces is not None else None
        self._non_stackable_edges = non_stackable_edges if isinstance(non_stackable_edges, np.ndarray) else np.array(non_stackable_edges, dtype=np.int32) if non_stackable_edges is not None else None
        self._small_features = small_features if isinstance(small_features, np.ndarray) else np.array(small_features, dtype=np.int32) if small_features is not None else None
        self._base_face_ids = base_face_ids if isinstance(base_face_ids, np.ndarray) else np.array(base_face_ids, dtype=np.int32) if base_face_ids is not None else None
        self._size_control_ids = size_control_ids if isinstance(size_control_ids, np.ndarray) else np.array(size_control_ids, dtype=np.int32) if size_control_ids is not None else None

    def __init__(
            self,
            model: CommunicationManager=None,
            error_code: ErrorCode = None,
            error_codes_per_topo_volume: Iterable[int] = None,
            non_stackable_faces: Iterable[int] = None,
            non_stackable_edges: Iterable[int] = None,
            small_features: Iterable[int] = None,
            base_face_ids: Iterable[int] = None,
            size_control_ids: Iterable[int] = None,
            json_data : dict = None,
             **kwargs):
        """Initializes the MeshStackerResults.

        Parameters
        ----------
        model: Model
            Model to create a MeshStackerResults object with default parameters.
        error_code: ErrorCode, optional
            Error code associated with the operation.
        error_codes_per_topo_volume: Iterable[int], optional
            Error codes associated with the topovolume-by-topovolume stacking.
        non_stackable_faces: Iterable[int], optional
            List of non-stackable faces.
        non_stackable_edges: Iterable[int], optional
            List of non-stackable edges.
        small_features: Iterable[int], optional
            List of features edges smaller than input tolerance.
        base_face_ids: Iterable[int], optional
            List of base face ids after base creation.
        size_control_ids: Iterable[int], optional
            List of ids of newly created size controls.
        json_data: dict, optional
            JSON dictionary to create a MeshStackerResults object with provided parameters.

        Examples
        --------
        >>> mesh_stacker_results = prime.MeshStackerResults(model = model)
        """
        if json_data:
            self.__initialize(
                ErrorCode(json_data["errorCode"] if "errorCode" in json_data else None),
                json_data["errorCodesPerTopoVolume"] if "errorCodesPerTopoVolume" in json_data else None,
                json_data["nonStackableFaces"] if "nonStackableFaces" in json_data else None,
                json_data["nonStackableEdges"] if "nonStackableEdges" in json_data else None,
                json_data["smallFeatures"] if "smallFeatures" in json_data else None,
                json_data["baseFaceIds"] if "baseFaceIds" in json_data else None,
                json_data["sizeControlIds"] if "sizeControlIds" in json_data else None)
        else:
            all_field_specified = all(arg is not None for arg in [error_code, error_codes_per_topo_volume, non_stackable_faces, non_stackable_edges, small_features, base_face_ids, size_control_ids])
            if all_field_specified:
                self.__initialize(
                    error_code,
                    error_codes_per_topo_volume,
                    non_stackable_faces,
                    non_stackable_edges,
                    small_features,
                    base_face_ids,
                    size_control_ids)
            else:
                if model is None:
                    raise ValueError("Invalid assignment. Either pass model or specify all properties")
                else:
                    param_json = model._communicator.initialize_params(model, "MeshStackerResults")
                    json_data = param_json["MeshStackerResults"] if "MeshStackerResults" in param_json else {}
                    self.__initialize(
                        error_code if error_code is not None else ( MeshStackerResults._default_params["error_code"] if "error_code" in MeshStackerResults._default_params else ErrorCode(json_data["errorCode"] if "errorCode" in json_data else None)),
                        error_codes_per_topo_volume if error_codes_per_topo_volume is not None else ( MeshStackerResults._default_params["error_codes_per_topo_volume"] if "error_codes_per_topo_volume" in MeshStackerResults._default_params else (json_data["errorCodesPerTopoVolume"] if "errorCodesPerTopoVolume" in json_data else None)),
                        non_stackable_faces if non_stackable_faces is not None else ( MeshStackerResults._default_params["non_stackable_faces"] if "non_stackable_faces" in MeshStackerResults._default_params else (json_data["nonStackableFaces"] if "nonStackableFaces" in json_data else None)),
                        non_stackable_edges if non_stackable_edges is not None else ( MeshStackerResults._default_params["non_stackable_edges"] if "non_stackable_edges" in MeshStackerResults._default_params else (json_data["nonStackableEdges"] if "nonStackableEdges" in json_data else None)),
                        small_features if small_features is not None else ( MeshStackerResults._default_params["small_features"] if "small_features" in MeshStackerResults._default_params else (json_data["smallFeatures"] if "smallFeatures" in json_data else None)),
                        base_face_ids if base_face_ids is not None else ( MeshStackerResults._default_params["base_face_ids"] if "base_face_ids" in MeshStackerResults._default_params else (json_data["baseFaceIds"] if "baseFaceIds" in json_data else None)),
                        size_control_ids if size_control_ids is not None else ( MeshStackerResults._default_params["size_control_ids"] if "size_control_ids" in MeshStackerResults._default_params else (json_data["sizeControlIds"] if "sizeControlIds" in json_data else None)))
        self._custom_params = kwargs
        if model is not None:
            [ model._logger.warning(f'Unsupported argument : {key}') for key in kwargs ]
        [setattr(type(self), key, property(lambda self, key = key:  self._custom_params[key] if key in self._custom_params else None,
        lambda self, value, key = key : self._custom_params.update({ key: value }))) for key in kwargs]
        self._freeze()

    @staticmethod
    def set_default(
            error_code: ErrorCode = None,
            error_codes_per_topo_volume: Iterable[int] = None,
            non_stackable_faces: Iterable[int] = None,
            non_stackable_edges: Iterable[int] = None,
            small_features: Iterable[int] = None,
            base_face_ids: Iterable[int] = None,
            size_control_ids: Iterable[int] = None):
        """Set the default values of MeshStackerResults.

        Parameters
        ----------
        error_code: ErrorCode, optional
            Error code associated with the operation.
        error_codes_per_topo_volume: Iterable[int], optional
            Error codes associated with the topovolume-by-topovolume stacking.
        non_stackable_faces: Iterable[int], optional
            List of non-stackable faces.
        non_stackable_edges: Iterable[int], optional
            List of non-stackable edges.
        small_features: Iterable[int], optional
            List of features edges smaller than input tolerance.
        base_face_ids: Iterable[int], optional
            List of base face ids after base creation.
        size_control_ids: Iterable[int], optional
            List of ids of newly created size controls.
        """
        args = locals()
        [MeshStackerResults._default_params.update({ key: value }) for key, value in args.items() if value is not None]

    @staticmethod
    def print_default():
        """Print the default values of MeshStackerResults.

        Examples
        --------
        >>> MeshStackerResults.print_default()
        """
        message = ""
        message += ''.join(str(key) + ' : ' + str(value) + '\n' for key, value in MeshStackerResults._default_params.items())
        print(message)

    def _jsonify(self) -> Dict[str, Any]:
        json_data = {}
        if self._error_code is not None:
            json_data["errorCode"] = self._error_code
        if self._error_codes_per_topo_volume is not None:
            json_data["errorCodesPerTopoVolume"] = self._error_codes_per_topo_volume
        if self._non_stackable_faces is not None:
            json_data["nonStackableFaces"] = self._non_stackable_faces
        if self._non_stackable_edges is not None:
            json_data["nonStackableEdges"] = self._non_stackable_edges
        if self._small_features is not None:
            json_data["smallFeatures"] = self._small_features
        if self._base_face_ids is not None:
            json_data["baseFaceIds"] = self._base_face_ids
        if self._size_control_ids is not None:
            json_data["sizeControlIds"] = self._size_control_ids
        [ json_data.update({ utils.to_camel_case(key) : value }) for key, value in self._custom_params.items()]
        return json_data

    def __str__(self) -> str:
        message = "error_code :  %s\nerror_codes_per_topo_volume :  %s\nnon_stackable_faces :  %s\nnon_stackable_edges :  %s\nsmall_features :  %s\nbase_face_ids :  %s\nsize_control_ids :  %s" % (self._error_code, self._error_codes_per_topo_volume, self._non_stackable_faces, self._non_stackable_edges, self._small_features, self._base_face_ids, self._size_control_ids)
        message += ''.join('\n' + str(key) + ' : ' + str(value) for key, value in self._custom_params.items())
        return message

    @property
    def error_code(self) -> ErrorCode:
        """Error code associated with the operation.
        """
        return self._error_code

    @error_code.setter
    def error_code(self, value: ErrorCode):
        self._error_code = value

    @property
    def error_codes_per_topo_volume(self) -> Iterable[int]:
        """Error codes associated with the topovolume-by-topovolume stacking.
        """
        return self._error_codes_per_topo_volume

    @error_codes_per_topo_volume.setter
    def error_codes_per_topo_volume(self, value: Iterable[int]):
        self._error_codes_per_topo_volume = value

    @property
    def non_stackable_faces(self) -> Iterable[int]:
        """List of non-stackable faces.
        """
        return self._non_stackable_faces

    @non_stackable_faces.setter
    def non_stackable_faces(self, value: Iterable[int]):
        self._non_stackable_faces = value

    @property
    def non_stackable_edges(self) -> Iterable[int]:
        """List of non-stackable edges.
        """
        return self._non_stackable_edges

    @non_stackable_edges.setter
    def non_stackable_edges(self, value: Iterable[int]):
        self._non_stackable_edges = value

    @property
    def small_features(self) -> Iterable[int]:
        """List of features edges smaller than input tolerance.
        """
        return self._small_features

    @small_features.setter
    def small_features(self, value: Iterable[int]):
        self._small_features = value

    @property
    def base_face_ids(self) -> Iterable[int]:
        """List of base face ids after base creation.
        """
        return self._base_face_ids

    @base_face_ids.setter
    def base_face_ids(self, value: Iterable[int]):
        self._base_face_ids = value

    @property
    def size_control_ids(self) -> Iterable[int]:
        """List of ids of newly created size controls.
        """
        return self._size_control_ids

    @size_control_ids.setter
    def size_control_ids(self, value: Iterable[int]):
        self._size_control_ids = value

class MeshStackerParams(CoreObject):
    """Input parameters associated with the mesh stacker operations.
    """
    _default_params = {}

    def __initialize(
            self,
            origin: Iterable[float],
            direction: Iterable[float],
            lateral_defeature_tolerance: float,
            stacking_defeature_tolerance: float,
            max_offset_size: float,
            size_control_ids: Iterable[int],
            delete_base: bool):
        self._origin = origin if isinstance(origin, np.ndarray) else np.array(origin, dtype=np.double) if origin is not None else None
        self._direction = direction if isinstance(direction, np.ndarray) else np.array(direction, dtype=np.double) if direction is not None else None
        self._lateral_defeature_tolerance = lateral_defeature_tolerance
        self._stacking_defeature_tolerance = stacking_defeature_tolerance
        self._max_offset_size = max_offset_size
        self._size_control_ids = size_control_ids if isinstance(size_control_ids, np.ndarray) else np.array(size_control_ids, dtype=np.int32) if size_control_ids is not None else None
        self._delete_base = delete_base

    def __init__(
            self,
            model: CommunicationManager=None,
            origin: Iterable[float] = None,
            direction: Iterable[float] = None,
            lateral_defeature_tolerance: float = None,
            stacking_defeature_tolerance: float = None,
            max_offset_size: float = None,
            size_control_ids: Iterable[int] = None,
            delete_base: bool = None,
            json_data : dict = None,
             **kwargs):
        """Initializes the MeshStackerParams.

        Parameters
        ----------
        model: Model
            Model to create a MeshStackerParams object with default parameters.
        origin: Iterable[float], optional
            Origin coordinate list of stacker.
        direction: Iterable[float], optional
            Direction vector of stacker.
        lateral_defeature_tolerance: float, optional
            Absolute lateral distance tolerance for stacker. Default value is 0.0.
        stacking_defeature_tolerance: float, optional
            Absolute stacking distance tolerance for stacker. Default value is 0.0.
        max_offset_size: float, optional
            Maximum stack size allowed during stacking.
        size_control_ids: Iterable[int], optional
            List of size control ids to be respected by stacker. Stacker respects all supported controls by default.
        delete_base: bool, optional
            Option to delete base face at the end of stacking. The default is false.
        json_data: dict, optional
            JSON dictionary to create a MeshStackerParams object with provided parameters.

        Examples
        --------
        >>> mesh_stacker_params = prime.MeshStackerParams(model = model)
        """
        if json_data:
            self.__initialize(
                json_data["origin"] if "origin" in json_data else None,
                json_data["direction"] if "direction" in json_data else None,
                json_data["lateralDefeatureTolerance"] if "lateralDefeatureTolerance" in json_data else None,
                json_data["stackingDefeatureTolerance"] if "stackingDefeatureTolerance" in json_data else None,
                json_data["maxOffsetSize"] if "maxOffsetSize" in json_data else None,
                json_data["sizeControlIds"] if "sizeControlIds" in json_data else None,
                json_data["deleteBase"] if "deleteBase" in json_data else None)
        else:
            all_field_specified = all(arg is not None for arg in [origin, direction, lateral_defeature_tolerance, stacking_defeature_tolerance, max_offset_size, size_control_ids, delete_base])
            if all_field_specified:
                self.__initialize(
                    origin,
                    direction,
                    lateral_defeature_tolerance,
                    stacking_defeature_tolerance,
                    max_offset_size,
                    size_control_ids,
                    delete_base)
            else:
                if model is None:
                    raise ValueError("Invalid assignment. Either pass model or specify all properties")
                else:
                    param_json = model._communicator.initialize_params(model, "MeshStackerParams")
                    json_data = param_json["MeshStackerParams"] if "MeshStackerParams" in param_json else {}
                    self.__initialize(
                        origin if origin is not None else ( MeshStackerParams._default_params["origin"] if "origin" in MeshStackerParams._default_params else (json_data["origin"] if "origin" in json_data else None)),
                        direction if direction is not None else ( MeshStackerParams._default_params["direction"] if "direction" in MeshStackerParams._default_params else (json_data["direction"] if "direction" in json_data else None)),
                        lateral_defeature_tolerance if lateral_defeature_tolerance is not None else ( MeshStackerParams._default_params["lateral_defeature_tolerance"] if "lateral_defeature_tolerance" in MeshStackerParams._default_params else (json_data["lateralDefeatureTolerance"] if "lateralDefeatureTolerance" in json_data else None)),
                        stacking_defeature_tolerance if stacking_defeature_tolerance is not None else ( MeshStackerParams._default_params["stacking_defeature_tolerance"] if "stacking_defeature_tolerance" in MeshStackerParams._default_params else (json_data["stackingDefeatureTolerance"] if "stackingDefeatureTolerance" in json_data else None)),
                        max_offset_size if max_offset_size is not None else ( MeshStackerParams._default_params["max_offset_size"] if "max_offset_size" in MeshStackerParams._default_params else (json_data["maxOffsetSize"] if "maxOffsetSize" in json_data else None)),
                        size_control_ids if size_control_ids is not None else ( MeshStackerParams._default_params["size_control_ids"] if "size_control_ids" in MeshStackerParams._default_params else (json_data["sizeControlIds"] if "sizeControlIds" in json_data else None)),
                        delete_base if delete_base is not None else ( MeshStackerParams._default_params["delete_base"] if "delete_base" in MeshStackerParams._default_params else (json_data["deleteBase"] if "deleteBase" in json_data else None)))
        self._custom_params = kwargs
        if model is not None:
            [ model._logger.warning(f'Unsupported argument : {key}') for key in kwargs ]
        [setattr(type(self), key, property(lambda self, key = key:  self._custom_params[key] if key in self._custom_params else None,
        lambda self, value, key = key : self._custom_params.update({ key: value }))) for key in kwargs]
        self._freeze()

    @staticmethod
    def set_default(
            origin: Iterable[float] = None,
            direction: Iterable[float] = None,
            lateral_defeature_tolerance: float = None,
            stacking_defeature_tolerance: float = None,
            max_offset_size: float = None,
            size_control_ids: Iterable[int] = None,
            delete_base: bool = None):
        """Set the default values of MeshStackerParams.

        Parameters
        ----------
        origin: Iterable[float], optional
            Origin coordinate list of stacker.
        direction: Iterable[float], optional
            Direction vector of stacker.
        lateral_defeature_tolerance: float, optional
            Absolute lateral distance tolerance for stacker. Default value is 0.0.
        stacking_defeature_tolerance: float, optional
            Absolute stacking distance tolerance for stacker. Default value is 0.0.
        max_offset_size: float, optional
            Maximum stack size allowed during stacking.
        size_control_ids: Iterable[int], optional
            List of size control ids to be respected by stacker. Stacker respects all supported controls by default.
        delete_base: bool, optional
            Option to delete base face at the end of stacking. The default is false.
        """
        args = locals()
        [MeshStackerParams._default_params.update({ key: value }) for key, value in args.items() if value is not None]

    @staticmethod
    def print_default():
        """Print the default values of MeshStackerParams.

        Examples
        --------
        >>> MeshStackerParams.print_default()
        """
        message = ""
        message += ''.join(str(key) + ' : ' + str(value) + '\n' for key, value in MeshStackerParams._default_params.items())
        print(message)

    def _jsonify(self) -> Dict[str, Any]:
        json_data = {}
        if self._origin is not None:
            json_data["origin"] = self._origin
        if self._direction is not None:
            json_data["direction"] = self._direction
        if self._lateral_defeature_tolerance is not None:
            json_data["lateralDefeatureTolerance"] = self._lateral_defeature_tolerance
        if self._stacking_defeature_tolerance is not None:
            json_data["stackingDefeatureTolerance"] = self._stacking_defeature_tolerance
        if self._max_offset_size is not None:
            json_data["maxOffsetSize"] = self._max_offset_size
        if self._size_control_ids is not None:
            json_data["sizeControlIds"] = self._size_control_ids
        if self._delete_base is not None:
            json_data["deleteBase"] = self._delete_base
        [ json_data.update({ utils.to_camel_case(key) : value }) for key, value in self._custom_params.items()]
        return json_data

    def __str__(self) -> str:
        message = "origin :  %s\ndirection :  %s\nlateral_defeature_tolerance :  %s\nstacking_defeature_tolerance :  %s\nmax_offset_size :  %s\nsize_control_ids :  %s\ndelete_base :  %s" % (self._origin, self._direction, self._lateral_defeature_tolerance, self._stacking_defeature_tolerance, self._max_offset_size, self._size_control_ids, self._delete_base)
        message += ''.join('\n' + str(key) + ' : ' + str(value) for key, value in self._custom_params.items())
        return message

    @property
    def origin(self) -> Iterable[float]:
        """Origin coordinate list of stacker.
        """
        return self._origin

    @origin.setter
    def origin(self, value: Iterable[float]):
        self._origin = value

    @property
    def direction(self) -> Iterable[float]:
        """Direction vector of stacker.
        """
        return self._direction

    @direction.setter
    def direction(self, value: Iterable[float]):
        self._direction = value

    @property
    def lateral_defeature_tolerance(self) -> float:
        """Absolute lateral distance tolerance for stacker. Default value is 0.0.
        """
        return self._lateral_defeature_tolerance

    @lateral_defeature_tolerance.setter
    def lateral_defeature_tolerance(self, value: float):
        self._lateral_defeature_tolerance = value

    @property
    def stacking_defeature_tolerance(self) -> float:
        """Absolute stacking distance tolerance for stacker. Default value is 0.0.
        """
        return self._stacking_defeature_tolerance

    @stacking_defeature_tolerance.setter
    def stacking_defeature_tolerance(self, value: float):
        self._stacking_defeature_tolerance = value

    @property
    def max_offset_size(self) -> float:
        """Maximum stack size allowed during stacking.
        """
        return self._max_offset_size

    @max_offset_size.setter
    def max_offset_size(self, value: float):
        self._max_offset_size = value

    @property
    def size_control_ids(self) -> Iterable[int]:
        """List of size control ids to be respected by stacker. Stacker respects all supported controls by default.
        """
        return self._size_control_ids

    @size_control_ids.setter
    def size_control_ids(self, value: Iterable[int]):
        self._size_control_ids = value

    @property
    def delete_base(self) -> bool:
        """Option to delete base face at the end of stacking. The default is false.
        """
        return self._delete_base

    @delete_base.setter
    def delete_base(self, value: bool):
        self._delete_base = value



""" Auto-generated file. DO NOT MODIFY """
from __future__ import annotations
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import *
from ansys.meshing.prime.autogen.coreobject import *
from typing import List, Any, Union

class BoundaryFittedSpline(CoreObject):
    """BoundaryFittedSpline helps you to create splines for structured hex-mesh model.

    BoundaryFittedSpline allows you to perform H and P refinement.
    """

    def __init__(self, model: CommunicationManager):
        """ Initialize BoundaryFittedSpline """
        self._model = model
        self._comm = model._communicator
        command_name = "PrimeMesh::BoundaryFittedSpline/Construct"
        args = {"ModelID" : model._object_id , "MaxID" : -1 }
        result = self._comm.serve(model, command_name, args=args)
        self._object_id = result["ObjectIndex"]
        self._freeze()

    def __enter__(self):
        """ Enter context for BoundaryFittedSpline. """
        return self

    def __exit__(self, type, value, traceback) :
        """ Exit context for BoundaryFittedSpline. """
        command_name = "PrimeMesh::BoundaryFittedSpline/Destruct"
        self._comm.serve(self._model, command_name, self._object_id, args={})

    def create_boundary_fitted_spline(self, part_id : int, cell_zonelet_ids : Iterable[int], boundary_fitted_spline_params : BoundaryFittedSplineParams) -> IGAResults:
        """ Create boundary fitted spline for structured hex-mesh.

        The hex-mesh can be structured in blocks but must be conformally connected.
        That is, each block must have six sided volume and must be connected to other blocks through unique face.
        The degree and number of control points of the spline can be set in the fitting parameter structure.

        Parameters
        ----------
        part_id : int
            Id of the part.
        cell_zonelet_ids : Iterable[int]
            Ids of the cell zonelets on which spline will be fit.
        boundary_fitted_spline_params : BoundaryFittedSplineParams
            Structure containing spline fitting parameters.

        Returns
        -------
        IGAResults
            Returns the IGAResults.


        Examples
        --------
        >>> from ansys.meshing.prime import BoundaryFittedSpline
        >>> #connect client to server and get model from it
        >>> client = Client(ip="localhost", port=50060)
        >>> model = client.model
        >>> boundary_fitted_spline = BoundaryFittedSpline(model = model)
        >>> results = boundary_fitted_spline.create_boundary_fitted_spline(part_id, cell_zonelet_ids, boundary_fitted_spline_params)

        """
        if not isinstance(part_id, int):
            raise TypeError("Invalid argument type passed for part_id, valid argument type is int.")
        if not isinstance(cell_zonelet_ids, Iterable):
            raise TypeError("Invalid argument type passed for cell_zonelet_ids, valid argument type is Iterable[int].")
        if not isinstance(boundary_fitted_spline_params, BoundaryFittedSplineParams):
            raise TypeError("Invalid argument type passed for boundary_fitted_spline_params, valid argument type is BoundaryFittedSplineParams.")
        args = {"part_id" : part_id,
        "cell_zonelet_ids" : cell_zonelet_ids,
        "boundary_fitted_spline_params" : boundary_fitted_spline_params._jsonify()}
        command_name = "PrimeMesh::BoundaryFittedSpline/CreateBoundaryFittedSpline"
        self._model._print_logs_before_command("create_boundary_fitted_spline", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("create_boundary_fitted_spline", IGAResults(model = self._model, json_data = result))
        return IGAResults(model = self._model, json_data = result)

    def refine_spline(self, part_id : int, spline_ids : Iterable[int], refine_spline_params : RefineSplineParams) -> IGAResults:
        """ Refine boundary fitted splines.

        Now H and P refinement are supported.
        Refinement along one or more dimension can be suppressed using refinement parameters in the input.

        Parameters
        ----------
        part_id : int
            Id of the part.
        spline_ids : Iterable[int]
            Ids of the splines on which refinement is performed.
        refine_spline_params : RefineSplineParams
            Structure containing parameters for spline refinement.

        Returns
        -------
        IGAResults
            Returns the IGAResults Structure.


        Examples
        --------
        >>> from ansys.meshing.prime import BoundaryFittedSpline
        >>> #connect client to server and get model from it
        >>> client = Client(ip="localhost", port=50060)
        >>> model = client.model
        >>> boundary_fitted_spline = BoundaryFittedSpline(model = model)
        >>> results = boundary_fitted_spline.refine_spline(part_id, spline_ids, refine_spline_params)

        """
        if not isinstance(part_id, int):
            raise TypeError("Invalid argument type passed for part_id, valid argument type is int.")
        if not isinstance(spline_ids, Iterable):
            raise TypeError("Invalid argument type passed for spline_ids, valid argument type is Iterable[int].")
        if not isinstance(refine_spline_params, RefineSplineParams):
            raise TypeError("Invalid argument type passed for refine_spline_params, valid argument type is RefineSplineParams.")
        args = {"part_id" : part_id,
        "spline_ids" : spline_ids,
        "refine_spline_params" : refine_spline_params._jsonify()}
        command_name = "PrimeMesh::BoundaryFittedSpline/RefineSpline"
        self._model._print_logs_before_command("refine_spline", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("refine_spline", IGAResults(model = self._model, json_data = result))
        return IGAResults(model = self._model, json_data = result)



"""Configuration utility for PyPrimeMesh."""

from contextlib import contextmanager

__all__ = [
    'enable_optimizing_numpy_arrays',
    'disable_optimizing_numpy_arrays',
    'is_optimizing_numpy_arrays',
    'numpy_array_optimization_enabled',
    'numpy_array_optimization_disabled',
    'set_using_container',
    'using_container',
    'set_has_pim',
    'has_pim',
    'enable_log_output',
]

__DEFAULT_USE_BINARY = False
__USING_CONTAINER = False
__HAS_PIM = False
__FILE_CHECK = True

from ansys.meshing.prime.internals.logger import PrimeLogger


def _optimize_vectors():
    """Get the value of the flag for optimizing vectors."""
    return __DEFAULT_USE_BINARY


def _set_optimize_vectors(optimize_vectors: bool):
    """Set the flag for optimizing vectors.

    Parameters
    ----------
    optimize_vectors : bool
        Whether to optimize vectors.

    Raises
    ------
    ValueError
        Value is not Boolean.
    """
    global __DEFAULT_USE_BINARY
    if not isinstance(optimize_vectors, bool):
        raise ValueError(f'Value to set optimize vectors flag should be Boolean.')

    __DEFAULT_USE_BINARY = optimize_vectors


def enable_optimizing_numpy_arrays():
    """Enable optimizing numpy arrays over the wire.

    This method allows the library to optimize serialization of numpy arrays
    for faster deserialization. This can cause the data transferred
    over the network to increase slightly unless the array is of a very
    large size. Consider enabling this option to improve performance if
    your workflow is not limited by network bandwidth for performance. The
    default for optimization is always ``True``.
    """
    _set_optimize_vectors(True)


def disable_optimizing_numpy_arrays():
    """Disable optimizing numpy arrays over the wire.

    This method disables serialization optimization for numpy arrays when
    sending them over the wire. Typically, this is needed in case the
    performance is limited by a saturated network. Disabling this flag
    can provide some benefit for data transferred over the wire for
    most arrays. The default for optimization is always ``True``.
    """
    _set_optimize_vectors(False)


def is_optimizing_numpy_arrays():
    """Determine if serialization of numpy arrays is enabled.

    The default for optimization is always ``True``.

    Returns
    -------
    bool
        Boolean flag indicating if optimizing of numpy arrays is enabled.
    """
    return _optimize_vectors()


@contextmanager
def numpy_array_optimization_enabled():
    """Context helper to execute code with numpy optimization enabled."""
    old_value = _optimize_vectors()
    _set_optimize_vectors(True)
    yield
    _set_optimize_vectors(old_value)


@contextmanager
def numpy_array_optimization_disabled():
    """Context helper to execute code with numpy optimization disabled."""
    old_value = _optimize_vectors()
    _set_optimize_vectors(False)
    yield
    _set_optimize_vectors(old_value)


def set_using_container(value: bool):
    """Set the ``USING_CONTAINER`` flag to ``value``.

    Parameters
    ----------
    value : bool
        Whether to use a container.

    Returns
    -------
    bool
        Value for the ``USING_CONTAINER`` flag.
    """
    global __USING_CONTAINER
    __USING_CONTAINER = value
    return __USING_CONTAINER


def using_container():
    """Get the value for the ``USING_CONTAINER`` flag.

    Returns
    -------
    bool
        Value for the ``USING_CONTAINER`` flag.
    """
    return __USING_CONTAINER


def set_has_pim(value: bool):
    """Set the `HAS_PIM` flag.

    PIM (Product Instance Manager) provides a gRPC API that
    enables both library and app developers to start a product in
    a remote environment and communicate with its API.

    Parameters
    ----------
    value : bool
        Value for setting the ``HAS_PIM`` flag.

    Returns
    -------
    bool
        Value for the ``HAS_PIM`` flag.
    """
    global __HAS_PIM
    __HAS_PIM = value
    return __HAS_PIM


def has_pim():
    """Get the ``HAS_PIM`` flag.

    Returns
    -------
    bool
        Value for the ``HAS_PIM`` flag.
    """
    return __HAS_PIM


def file_existence_check_enabled():
    """Get config to enable checking if files exist.

    Returns
    -------
    bool
        Option to check if files exist before reading them.
    """
    return __FILE_CHECK


def set_file_existence_check(value: bool):
    """Set option to enable checking if files exist.

    Parameters
    ----------
    value : bool
        Value to set the option with.

    Returns
    -------
    bool
        Option to check if files exist before reading them.
    """
    global __FILE_CHECK
    __FILE_CHECK = value
    return __FILE_CHECK


def enable_log_output(stream=None):
    """Enable logger output to given stream.

    If stream is not specified, sys.stderr is used.

    Parameters
    ----------
    stream: TextIO, optional
        Stream to output the log output to stream
    """
    PrimeLogger().enable_output(stream)



"""Implementation of Relaxed JSON Decoder."""
import re
from typing import Any, Callable, Dict, Tuple

import numpy as np

from . import scanner

__all__ = ['JSONDecoder', 'JSONDecodeError']

FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL

NaN = float('nan')
PosInf = float('inf')
NegInf = float('-inf')

_CONSTANTS = {
    b'-Infinity': NegInf,
    b'Infinity': PosInf,
    b'NaN': NaN,
}


class JSONDecodeError(ValueError):
    """Errors for JSONDecode class.

    Parameters
    ----------
    msg : str
        Error message
    document : bytesorbytearray
        Document where the error is happening.
    pos : int
        Position where the error is happening.
    """

    def __init__(self, msg: str, document: bytes or bytearray, pos: int):
        """Initialize error handler."""
        lineno = document.count(b'\n', 0, pos) + 1
        colno = pos - document.rfind(b'\n', 0, pos)
        error_msg = f"{msg}: line {lineno} column {colno} (char {pos})"
        ValueError.__init__(self, error_msg)
        self.error_msg = error_msg
        self.lineno = lineno
        self.colno = colno
        self.pos = pos


WHITESPACE = re.compile(r'[ \t\n\r]*'.encode("utf-8"), FLAGS)
WHITESPACE_STR = b' \t\n\r'

STRINGCHUNK = re.compile(r'(.*?)(["\\])'.encode('utf-8'), FLAGS)
BACKSLASH = {
    '"': b'"',
    '\\': b'\\',
    '/': b'/',
    'b': b'\b',
    'f': b'\f',
    'n': b'\n',
    'r': b'\r',
    't': b'\t',
}


def then(first, continuation):
    """Iterate function."""

    def _then(dct):
        dct = first(dct)
        return continuation(dct)

    return _then


TYPEMAP = {
    0x11: np.uint8,
    0x12: np.int16,
    0x13: np.int32,
    0x14: np.int64,
    0x19: np.uint8,
    0x1A: np.uint16,
    0x1B: np.uint32,
    0x1C: np.uint64,
    0x21: np.float32,
    0x22: np.float64,
}

TYPEDATA = {
    'int32': {'type': int, 'object_size': 4, 'signed': True},
    'int64': {'type': int, 'object_size': 8, 'signed': True},
    'uint32': {'type': int, 'object_size': 4, 'signed': False},
    'uint64': {'type': int, 'object_size': 8, 'signed': True},
    'double': {'type': float, 'object_size': 8},
}


def scan_string(str: bytes or bytearray, end: int, _b=BACKSLASH, _m=STRINGCHUNK.match):
    """Scan the string s for a JSON string.

    Scan the string s for a JSON string. End is the index of the
    character in s after the quote that started the JSON string.
    Unescapes all valid JSON string escape sequences and raises ValueError
    on attempt to decode an invalid string. If strict is False then literal
    control characters are allowed in the string.

    Returns a tuple of the decoded string and the index of the character in s
    after the end quote.
    """
    chunks = bytearray()
    begin = end - 1
    while 1:
        chunk = _m(str, end)
        if chunk is None:
            raise JSONDecodeError("Unterminated string starting at", str, begin)
        end = chunk.end()
        content, terminator = chunk.groups()
        # Content is contains zero or more unescaped string characters
        if content:
            chunks += content
        # Terminator is the end of string, a literal control character,
        # or a backslash denoting that an escape sequence follows
        if terminator == b'"':
            break
        elif terminator != b'\\':
            chunks += terminator
            continue
        try:
            esc = str[end : end + 1]
        except IndexError:
            raise JSONDecodeError("Unterminated string starting at", str, begin) from None
        # If not a unicode escape sequence, must be in the lookup table
        if esc == b'u':
            raise JSONDecodeError("Unicode not supported", str, end)
        else:
            try:
                char: bytes or bytearray = _b[esc.decode('utf-8')]
            except KeyError:
                msg = "Invalid \\escape: {0!r}".format(esc.decode('utf-8'))
                raise JSONDecodeError(msg, str, end)
            end += 1
        chunks += char

    try:
        chunks = chunks.decode('utf-8')
    except:
        raise JSONDecodeError("String needs to be utf-8 encoded", str, begin)
    return chunks, end


def decode_bytes(s, end):
    """Decode bytes object to JSON object.

    Parameters
    ----------
    s : Any
        Object to decode.
    end : int
        Final position.

    Returns
    -------
    np.array
        Array with decoded values.

    Raises
    ------
    JSONDecodeError
        Invalid values.
    """
    begin = end - 1
    size = int.from_bytes(bytes(s[end : end + 8]), byteorder='big')
    end += 8
    type = int.from_bytes(s[end : end + 1], byteorder='big')
    end += 1
    value = bytes(s[end : end + size])
    end += size

    if s[end : end + 1] != b'>':
        raise JSONDecodeError("Unterminated bytes object starting at", str, begin) from None
    end += 1  # '>' character

    dtype = TYPEMAP.get(type, None)
    if dtype is None:
        raise JSONDecodeError(f'Invalid type ({type}) decoded while reading bytes', s, end)
    dtype = np.dtype(dtype)
    data = np.frombuffer(value, dtype=dtype.newbyteorder('>'))
    return np.array(data, dtype=dtype), end


def decode_object(
    str_and_idx: Tuple[str, int],
    scan_once: Callable[[str, int], Tuple[str, int]],
    object_hook: Callable[[Dict[str, Any]], Any],
    object_pairs_hook,
    memo: Dict,
    _w=WHITESPACE.match,
    _ws=WHITESPACE_STR,
):
    """Decode generic object to JSON object.

    Parameters
    ----------
    str_and_idx : Tuple[str, int]
        String and index to decode.
    scan_once : Callable[[str, int], Tuple[str, int]]
        Scanner for bytes and bytearray.
    object_hook : Callable[[Dict[str, Any]], Any]
        Objects that are related.
    object_pairs_hook : Any
        Objects that are related.
    memo : Dict
        Needed for backwards compatibility.
    _w : str, optional
        Whitespace regex, by default WHITESPACE.match.
    _ws : str, optional
        Whitespace string, by default WHITESPACE_STR.

    Returns
    -------
    Tuple
        Decoded pairs and end index.

    Raises
    ------
    JSONDecodeError
        Unexpected value.
    """
    s, end = str_and_idx
    pairs = []
    pairs_append = pairs.append
    # Backwards compatibility
    if memo is None:
        memo = {}
    memo_get = memo.setdefault
    # Use a slice to prevent IndexError from being raised, the following
    # check will raise a more specific ValueError if the string is empty
    nextchar = bytes(s[end : end + 1])
    # Normally we expect nextchar == '"'
    if nextchar != b'"':
        if nextchar in _ws:
            end = _w(s, end).end()
            nextchar = bytes(s[end : end + 1])
        # Trivial empty object
        if nextchar == b'}':
            if object_pairs_hook is not None:
                result = object_pairs_hook(pairs)
                return result, end + 1
            pairs = {}
            if object_hook is not None:
                pairs = object_hook(pairs)
            return pairs, end + 1
        elif nextchar != b'"':
            raise JSONDecodeError("Expecting property name enclosed in double quotes", s, end)
    end += 1
    while True:
        key, end = scan_string(s, end)
        key = memo_get(key, key)
        # To skip some function call overhead we optimize the fast paths where
        # the JSON key separator is ": " or just ":".
        if s[end : end + 1] != b':':
            end = _w(s, end).end()
            if s[end : end + 1] != b':':
                raise JSONDecodeError("Expecting ':' delimiter", s, end)
        end += 1

        try:
            if s[end : end + 1] in _ws:
                end += 1
                if s[end : end + 1] in _ws:
                    end = _w(s, end + 1).end()
        except IndexError:
            pass

        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        pairs_append((key, value))
        try:
            nextchar = bytes(s[end : end + 1])
            if nextchar in _ws:
                end = _w(s, end + 1).end()
                nextchar = bytes(s[end : end + 1])
        except IndexError:
            nextchar = b''
        end += 1

        if nextchar == b'}':
            break
        elif nextchar != b',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)
        end = _w(s, end).end()
        nextchar = s[end : end + 1]
        end += 1
        if nextchar != b'"':
            raise JSONDecodeError("Expecting property name enclosed in double quotes", s, end - 1)
    if object_pairs_hook is not None:
        result = object_pairs_hook(pairs)
        return result, end
    pairs = dict(pairs)
    if object_hook is not None:
        pairs = object_hook(pairs)
    return pairs, end


def decode_array(
    str_and_idx: Tuple[str, int],
    scan_once: Callable[[str, int], Tuple[str, int]],
    _w=WHITESPACE.match,
    _ws=WHITESPACE_STR,
):
    """Decode array to JSON object.

    Parameters
    ----------
    str_and_idx : Tuple[str, int]
        string and index to decode.
    scan_once : Callable[[str, int], Tuple[str, int]]
        Scanner for bytes and bytearray.
    _w : str, optional
        Whitespace regex, by default WHITESPACE.match.
    _ws : str, optional
        Whitespace string, by default WHITESPACE_STR.

    Returns
    -------
    Tuple
        Decoded values and end index.

    Raises
    ------
    JSONDecodeError
        Unexpected value.
    JSONDecodeError
        Unexpected delimiter.
    """
    s, end = str_and_idx
    values = []
    nextchar = bytes(s[end : end + 1])
    if nextchar in _ws:
        end = _w(s, end + 1).end()
        nextchar = s[end : end + 1]
    # Look-ahead for trivial empty array
    if nextchar == b']':
        return values, end + 1
    _append = values.append
    while True:
        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        _append(value)
        nextchar = bytes(s[end : end + 1])
        if nextchar in _ws:
            end = _w(s, end + 1).end()
            nextchar = s[end : end + 1]
        end += 1
        if nextchar == b']':
            break
        elif nextchar != b',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)
        try:
            if bytes(s[end : end + 1]) in _ws:
                end += 1
                if bytes(s[end : end + 1]) in _ws:
                    end = _w(s, end + 1).end()
        except IndexError:
            pass

    return values, end


class JSONDecoder:
    """Decode object to JSON format.

    Parameters
    ----------
    object_hook : Any, optional
        Unused parameter, by default None
    parse_float : String or float, optional
        Parses float values, by default None
    parse_int : String or int, optional
        Parses int values, by default None
    parse_constant : Any, optional
        Parses constant values, by default Noneor_CONSTANTS.__getitem__
    strict : bool, optional
        Unused parameter, by default True
    object_pairs_hook : Any, optional
        Unused parameter, by default None
    """

    def __init__(
        self,
        *,
        object_hook=None,
        parse_float=None,
        parse_int=None,
        parse_constant=None or _CONSTANTS.__getitem__,
        strict=True,
        object_pairs_hook=None,
    ):
        """Initialize JSON decoder."""
        self.object_hook = object_hook or None
        self.parse_float = parse_float or float
        self.parse_int = parse_int or int
        self.parse_constant = parse_constant
        self.object_pairs_hook = object_pairs_hook
        self.parse_string = scan_string
        self.parse_array = decode_array
        self.parse_object = decode_object
        self.parse_bytes = decode_bytes
        self.strict = strict
        self.memo = {}
        self.scan_once = scanner.make_scanner(self)

    def raw_decode(self, s: bytes or bytearray, idx: int) -> Tuple[Any, int]:
        """Decode bytes or bytearray to a JSON object.

        Parameters
        ----------
        s : bytesorbytearray
            Information to decode.
        idx : int
            Position to decode.

        Returns
        -------
        Tuple[Any, int]
            JSON object and final position.

        Raises
        ------
        JSONDecodeError
            Unexpected value.
        """
        try:
            obj, end = self.scan_once(s, idx)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        return obj, end

    def decode(self, s: bytes or bytearray, _w=WHITESPACE.match) -> Any:
        """Decode bytes or bytearray to a JSON object.

        Parameters
        ----------
        s : bytesorbytearray
            Information to decode.
        _w : str, optional
            White space regex, by default WHITESPACE.match.

        Returns
        -------
        Any
            JSON object.

        Raises
        ------
        JSONDecodeError
            Document malformed.
        """
        obj, end = self.raw_decode(s, _w(s, 0).end())
        end = _w(s, end).end()
        if end != len(s):
            raise JSONDecodeError("Extra data at end of document", s, end)
        return obj.decode('utf-8') if isinstance(obj, (bytes, bytearray)) else obj



""" Auto-generated file. DO NOT MODIFY """
from __future__ import annotations
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import *
from ansys.meshing.prime.autogen.coreobject import *
from typing import List, Any, Union

class Morpher(CoreObject):
    """Morpher contain functionalities to change the geometry, adapting the mesh accordingly without a recomputation.

    """

    def __init__(self, model: CommunicationManager):
        """ Initialize Morpher """
        self._model = model
        self._comm = model._communicator
        command_name = "PrimeMesh::Morpher/Construct"
        args = {"ModelID" : model._object_id , "MaxID" : -1 }
        result = self._comm.serve(model, command_name, args=args)
        self._object_id = result["ObjectIndex"]
        self._freeze()

    def __enter__(self):
        """ Enter context for Morpher. """
        return self

    def __exit__(self, type, value, traceback) :
        """ Exit context for Morpher. """
        command_name = "PrimeMesh::Morpher/Destruct"
        self._comm.serve(self._model, command_name, self._object_id, args={})

    def match_morph(self, part_id : int, match_pairs : List[MatchPair], match_morph_params : MatchMorphParams, bc_params : MorphBCParams, solve_params : MorphSolveParams) -> MatchMorphResults:
        """ Match source and target zonelets defined by match pairs with prescribed boundary conditions for each pair. Also, solves boundary condition parameters to define uniform surface and volume regions in source neighborhood.


        Parameters
        ----------
        part_id : int
            Id of source part.
        match_pairs : MatchPairArray
            Array of match pairs of sources and targets.
        match_morph_params : MatchMorphParams
            Match morph parameters.
        bc_params : MorphBCParams
            Morph boundary condition parameters.
        solve_params : MorphSolveParams
            Morpher solve parameters.

        Returns
        -------
        MatchMorphResults
            Returns the MatchMorphResults.


        Examples
        --------
        >>> result = morph.match_morph([match_pair1, match_pair2], match_morph_params, bc_params, solve_params)

        """
        if not isinstance(part_id, int):
            raise TypeError("Invalid argument type passed for part_id, valid argument type is int.")
        if not isinstance(match_pairs, List):
            raise TypeError("Invalid argument type passed for match_pairs, valid argument type is List[MatchPair].")
        if not isinstance(match_morph_params, MatchMorphParams):
            raise TypeError("Invalid argument type passed for match_morph_params, valid argument type is MatchMorphParams.")
        if not isinstance(bc_params, MorphBCParams):
            raise TypeError("Invalid argument type passed for bc_params, valid argument type is MorphBCParams.")
        if not isinstance(solve_params, MorphSolveParams):
            raise TypeError("Invalid argument type passed for solve_params, valid argument type is MorphSolveParams.")
        args = {"part_id" : part_id,
        "match_pairs" : [p._jsonify() for p in match_pairs],
        "match_morph_params" : match_morph_params._jsonify(),
        "bc_params" : bc_params._jsonify(),
        "solve_params" : solve_params._jsonify()}
        command_name = "PrimeMesh::Morpher/MatchMorph"
        self._model._print_logs_before_command("match_morph", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("match_morph", MatchMorphResults(model = self._model, json_data = result))
        return MatchMorphResults(model = self._model, json_data = result)



"""Module for control data."""
from typing import Iterable, List

# isort: split
from ansys.meshing.prime.autogen.controldata import ControlData as _ControlData

# isort: split
from ansys.meshing.prime.autogen.commonstructs import DeleteResults
from ansys.meshing.prime.autogen.primeconfig import ErrorCode
from ansys.meshing.prime.autogen.prismcontrol import PrismControl
from ansys.meshing.prime.autogen.thinvolumecontrol import ThinVolumeControl
from ansys.meshing.prime.core.multizonecontrol import MultiZoneControl
from ansys.meshing.prime.core.periodiccontrol import PeriodicControl
from ansys.meshing.prime.core.sizecontrol import SizeControl
from ansys.meshing.prime.core.volumecontrol import VolumeControl
from ansys.meshing.prime.core.wrappercontrol import WrapperControl
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import SizingType


class ControlData(_ControlData):
    """Contains all controls.

    This class contains all controls, including size controls, prism controls, multizone controls
    and wrapper controls.

    Parameters
    ----------
    model : CommunicationManager
        Communicator with the server.
    id : int
        ID of the model.
    object_id : int
        ID of the object.
    name : str
        Name of the control.

    """

    def __init__(self, model: CommunicationManager, id: int, object_id: int, name: str):
        """Initialize the ``ControlData`` class."""
        self._model = model
        self._wrapper_controls = []
        self._mz_controls = []
        self._size_controls = []
        self._prism_controls = []
        self._thin_volume_controls = []
        self._volume_controls = []
        self._periodic_controls = []
        _ControlData.__init__(self, model, id, object_id, name)

    def get_wrapper_control_by_name(self, name) -> WrapperControl:
        """Get the wrapper control by name.

        Parameters
        ----------
        name : str
            Name of the wrapper control.

        Returns
        -------
        WrapperControl
            Wrapper control.

        Examples
        --------
        >>> wrapper_control = model.control_data.get_wrapper_control_by_name("wrappercontrol-1")

        """
        for wc in self._wrapper_controls:
            if wc.name == name:
                return wc
        return None

    def get_multi_zone_control_by_name(self, name) -> MultiZoneControl:
        """Get the multizone control by name.

        Parameters
        ------------
        name : str
            Name of the multizone control.

        Returns
        -------
        MultiZoneControl
            Returns the multizone control.

        Examples
        --------
        >>> multi_zone_control = model.control_data.get_multi_zone_control_by_name("mzcontrol-1")

        """
        for mc in self._mz_controls:
            if mc.name == name:
                return mc
        return None

    def create_size_control(self, sizing_type: SizingType) -> SizeControl:
        """Create a size control for a sizing type.

        Parameters
        ----------
        type : SizingType
            Sizing type for creating the size control.

        Returns
        -------
        SizeControl
            Size control.

        Notes
        -----
        An empty size control is created on calling this method.

        Examples
        --------
        >>> size_control = model.control_data.create_size_control(SizingType.Curvature)

        """
        res = _ControlData.create_size_control(self, sizing_type)
        new_size_control = SizeControl(self._model, res[0], res[1], res[2])
        self._size_controls.append(new_size_control)
        return new_size_control

    def create_prism_control(self) -> PrismControl:
        """Create a prism control.

        Returns
        -------
        PrismControl
            Prism control.


        Examples
        --------
        >>> prism_control = model.control_data.create_prism_control()

        """
        res = _ControlData.create_prism_control(self)
        new_prism_control = PrismControl(self._model, res[0], res[1], res[2])
        self._prism_controls.append(new_prism_control)
        return new_prism_control

    def create_thin_volume_control(self) -> ThinVolumeControl:
        """Create a thin volume control.

        Returns
        -------
        ThinVolumeControl
            Thin volume control.

        Examples
        --------
        >>> thin_volume_control = model.control_data.create_thin_volume_control()

        """
        res = _ControlData.create_thin_volume_control(self)
        new_thin_volume_control = ThinVolumeControl(self._model, res[0], res[1], res[2])
        self._thin_volume_controls.append(new_thin_volume_control)
        return new_thin_volume_control

    def create_wrapper_control(self) -> WrapperControl:
        """Create a wrapper control with default values.

        Returns
        -------
        WrapperControl
            Wrapper control.

        Notes
        -----
        A wrapper control with default values is created on calling this method.

        Examples
        --------
        >>> wrapper_control = model.control_data.create_wrapper_control()

        """
        res = _ControlData.create_wrapper_control(self)
        new_control = WrapperControl(self._model, res[0], res[1], res[2])
        self._wrapper_controls.append(new_control)
        return new_control

    def create_multi_zone_control(self) -> MultiZoneControl:
        """Create multizone control with defaults.

        Returns
        -------
        multizone
            Returns the multizone control.

        Examples
        --------
        >>> multizone = model.control_data.create_wrapper_control()

        """
        res = _ControlData.create_multi_zone_control(self)
        new_control = MultiZoneControl(self._model, res[0], res[1], res[2])
        self._mz_controls.append(new_control)
        return new_control

    def get_size_control_by_name(self, name: str) -> SizeControl:
        """Get a size control by name.

        Parameters
        ----------
        name : str
            Name of the size control.

        Returns
        -------
        SizeControl
            Size control.

        Examples
        --------
        >>> size_control = model.control_data.get_size_control_by_name("SizeControl-1")

        """
        for size_control in self._size_controls:
            if size_control.name == name:
                return size_control
        return None

    def get_prism_control_by_name(self, name: str) -> PrismControl:
        """Get a prism control by name.

        Parameters
        ----------
        name : str
            Name of the prism control.

        Returns
        -------
        PrismControl
            Prism control.

        Examples
        --------
        >>> prism_control = model.control_data.get_prism_control_by_name("PrismControl-1")

        """
        for prism_control in self._prism_controls:
            if prism_control.name == name:
                return prism_control
        return None

    def get_thin_volume_control_by_name(self, name: str) -> ThinVolumeControl:
        """Get a thin volume control by name.

        Parameters
        ----------
        name : str
            Name of the thin volume control.

        Returns
        -------
        ThinVolumeControl
            Thin volume control.

        Examples
        --------
        >>> contorl_data = model.control_data
        >>> thin_volume_control = control_data.get_thin_volume_control_by_name(
                                                     "ThinVolumeControl-1")
        """
        for thin_volume_control in self._thin_volume_controls:
            if thin_volume_control.name == name:
                return thin_volume_control
        return None

    def delete_controls(self, control_ids: Iterable[int]) -> DeleteResults:
        """Delete the control for one or more IDs.

        Parameters
        ----------
        control_ids : Iterable[int]
            List of control IDs.

        Returns
        -------
        DeleteResults
            Delete results.

        Examples
        --------
        >>> results = model.control_data.delete_controls([size_control.id, volume_control.id])

        """
        res = _ControlData.delete_controls(self, control_ids=control_ids)
        if res.error_code == ErrorCode.NOERROR:
            for id in control_ids:
                for size_control in self._size_controls:
                    if size_control.id == id:
                        self._size_controls.remove(size_control)
                        break
                for wrapper_control in self._wrapper_controls:
                    if wrapper_control.id == id:
                        self._wrapper_controls.remove(wrapper_control)
                        break
                for prism_control in self._prism_controls:
                    if prism_control.id == id:
                        self._prism_controls.remove(prism_control)
                        break
                for thin_volume_control in self._thin_volume_controls:
                    if thin_volume_control.id == id:
                        self._thin_volume_controls.remove(thin_volume_control)
                        break
                for volume_control in self._volume_controls:
                    if volume_control.id == id:
                        self._volume_controls.remove(volume_control)
                        break
                for periodic_control in self._periodic_controls:
                    if periodic_control.id == id:
                        self._periodic_controls.remove(periodic_control)
                        break
                for multi_zone_control in self._mz_controls:
                    if multi_zone_control.id == id:
                        self._mz_controls.remove(multi_zone_control)
                        break
        return res

    def _update_size_controls(self, c_data: List):
        self._size_controls = [SizeControl(self._model, c[0], c[1], c[2]) for c in c_data]

    def _update_prism_controls(self, c_data: List):
        self._prism_controls = [PrismControl(self._model, c[0], c[1], c[2]) for c in c_data]

    def _update_wrapper_controls(self, c_data: List):
        self._wrapper_controls = [WrapperControl(self._model, c[0], c[1], c[2]) for c in c_data]

    def _update_multi_zone_controls(self, c_data: List):
        self._mz_controls = [MultiZoneControl(self._model, c[0], c[1], c[2]) for c in c_data]

    def _update_volume_controls(self, c_data: List):
        self._volume_controls = [VolumeControl(self._model, c[0], c[1], c[2]) for c in c_data]

    def _update_periodic_controls(self, c_data: List):
        self._periodic_controls = [PeriodicControl(self._model, c[0], c[1], c[2]) for c in c_data]

    def _update_thin_volume_controls(self, c_data: List):
        self._thin_volume_controls = [
            ThinVolumeControl(self._model, c[0], c[1], c[2]) for c in c_data
        ]

    def create_volume_control(self) -> VolumeControl:
        """Create a volume control.

        Returns
        -------
        VolumeControl
            Volume control.

        Examples
        --------
        >>> volume_control = model.control_data.create_volume_control()

        """
        res = _ControlData.create_volume_control(self)
        new_control = VolumeControl(self._model, res[0], res[1], res[2])
        self._volume_controls.append(new_control)
        return new_control

    def get_volume_control_by_name(self, name: str) -> VolumeControl:
        """Get a volume control by name.

        Parameters
        ----------
        name : str
            Name of the volume control.

        Returns
        -------
        VolumeControl
            Volume control.

        Examples
        --------
        >>> volume_control = model.control_data.get_volume_control_by_name("VolumeControl-1")

        """
        for volume_control in self._volume_controls:
            if volume_control.name == name:
                return volume_control
        return None

    @property
    def size_controls(self) -> List[SizeControl]:
        """Get the size controls.

        Returns
        -------
        List[SizeControl]
            List of size controls.

        Examples
        --------
            >>> size_controls = model.control_data.size_controls
        """
        return self._size_controls

    @property
    def volume_controls(self) -> List[VolumeControl]:
        """Get the volume controls.

        Returns
        -------
        List[VolumeControl]
            List of volume controls.

        Examples
        --------
            >>> volume_controls = model.control_data.volume_controls
        """
        return self._volume_controls

    @property
    def prism_controls(self) -> List[PrismControl]:
        """Get the prism controls.

        Returns
        -------
        List[PrismControl]
            List of prism controls.

        Examples
        --------
        >>> prism_control = model.control_data.prism_controls

        """
        return self._prism_controls

    @property
    def thin_volume_controls(self) -> List[ThinVolumeControl]:
        """Get the thin volume controls.

        Returns
        -------
        List[ThinVolumeControl]
            List of thin volume controls.

        Examples
        --------
        >>> thin_volume_control = model.control_data.thin_volume_controls

        """
        return self._thin_volume_controls

    @property
    def wrapper_controls(self) -> List[WrapperControl]:
        """Get the wrapper controls.

        Returns
        -------
        List[WrapperControl]
            List of wrapper controls.

        Examples
        --------
        >>> wrapper_control = model.control_data.wrapper_controls

        """
        return self._wrapper_controls

    def multi_zone_controls(self) -> List[MultiZoneControl]:
        """Get the multizone controls.

        Returns
        -------
        List[MultiZoneControl]
            Returns the list of mutlizone controls.

        Examples
        --------
        >>> multi_zone_control = model.control_data.mz_controls

        """
        return self._mz_controls

    def create_periodic_control(self) -> PeriodicControl:
        """Create a periodic control.

        Returns
        -------
        PeriodicControl
            Periodic control.

        Examples
        --------
        >>> periodic_control = model.control_data.create_periodic_control()

        """
        res = _ControlData.create_periodic_control(self)
        new_control = PeriodicControl(self._model, res[0], res[1], res[2])
        self._periodic_controls.append(new_control)
        return new_control

    def get_periodic_control_by_name(self, name: str) -> PeriodicControl:
        """Get a periodic control by name.

        Parameters
        ----------
        name : str
            Name of the periodic control.

        Returns
        -------
        PeriodicControl
            Periodic control.

        Examples
        --------
        >>> periodic_control = model.control_data.get_periodic_control_by_name("PeriodicControl-1")

        """
        for periodic_control in self._periodic_controls:
            if periodic_control.name == name:
                return periodic_control
        return None

    @property
    def periodic_controls(self) -> List[PeriodicControl]:
        """Get the periodic controls.

        Returns
        -------
        List[PeriodicControl]
            List of periodic controls.

        Examples
        --------
            >>> periodic_controls = model.control_data.periodic_controls
        """
        return self._periodic_controls



""" Auto-generated file. DO NOT MODIFY """
import enum
from typing import Dict, Any, Union, List, Iterable
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.internals import utils
from ansys.meshing.prime.autogen.coreobject import *
import numpy as np

from ansys.meshing.prime.params.primestructs import *

class BCsVolumetricModality(enum.IntEnum):
    """Indicate options to identify morphable region in input volumetric mesh.
    """
    BOX = 1
    """Option to identify nodes enclosed by bounding box as morphable nodes from the input volumetric mesh."""
    ALL = 2
    """Option to identify all nodes expect fixed nodes as morphable nodes from the input volumetric mesh."""

class MorphBCParams(CoreObject):
    """MorphBCParams contains the input parameters for calculating the boundary conditions for a morphing problem.
    """
    _default_params = {}

    def __initialize(
            self,
            morph_region_method: BCsVolumetricModality,
            morphable_layers: int,
            morph_region_box_extension: float):
        self._morph_region_method = BCsVolumetricModality(morph_region_method)
        self._morphable_layers = morphable_layers
        self._morph_region_box_extension = morph_region_box_extension

    def __init__(
            self,
            model: CommunicationManager=None,
            morph_region_method: BCsVolumetricModality = None,
            morphable_layers: int = None,
            morph_region_box_extension: float = None,
            json_data : dict = None,
             **kwargs):
        """Initializes the MorphBCParams.

        Parameters
        ----------
        model: Model
            Model to create a MorphBCParams object with default parameters.
        morph_region_method: BCsVolumetricModality, optional
            Indicate options to identify morphable regions in the input volumetric mesh.
        morphable_layers: int, optional
            Number of layers around defined boundary.
        morph_region_box_extension: float, optional
            Percentage extension for the volumetric box.
        json_data: dict, optional
            JSON dictionary to create a MorphBCParams object with provided parameters.

        Examples
        --------
        >>> morph_bcparams = prime.MorphBCParams(model = model)
        """
        if json_data:
            self.__initialize(
                BCsVolumetricModality(json_data["morphRegionMethod"] if "morphRegionMethod" in json_data else None),
                json_data["morphableLayers"] if "morphableLayers" in json_data else None,
                json_data["morphRegionBoxExtension"] if "morphRegionBoxExtension" in json_data else None)
        else:
            all_field_specified = all(arg is not None for arg in [morph_region_method, morphable_layers, morph_region_box_extension])
            if all_field_specified:
                self.__initialize(
                    morph_region_method,
                    morphable_layers,
                    morph_region_box_extension)
            else:
                if model is None:
                    raise ValueError("Invalid assignment. Either pass model or specify all properties")
                else:
                    param_json = model._communicator.initialize_params(model, "MorphBCParams")
                    json_data = param_json["MorphBCParams"] if "MorphBCParams" in param_json else {}
                    self.__initialize(
                        morph_region_method if morph_region_method is not None else ( MorphBCParams._default_params["morph_region_method"] if "morph_region_method" in MorphBCParams._default_params else BCsVolumetricModality(json_data["morphRegionMethod"] if "morphRegionMethod" in json_data else None)),
                        morphable_layers if morphable_layers is not None else ( MorphBCParams._default_params["morphable_layers"] if "morphable_layers" in MorphBCParams._default_params else (json_data["morphableLayers"] if "morphableLayers" in json_data else None)),
                        morph_region_box_extension if morph_region_box_extension is not None else ( MorphBCParams._default_params["morph_region_box_extension"] if "morph_region_box_extension" in MorphBCParams._default_params else (json_data["morphRegionBoxExtension"] if "morphRegionBoxExtension" in json_data else None)))
        self._custom_params = kwargs
        if model is not None:
            [ model._logger.warning(f'Unsupported argument : {key}') for key in kwargs ]
        [setattr(type(self), key, property(lambda self, key = key:  self._custom_params[key] if key in self._custom_params else None,
        lambda self, value, key = key : self._custom_params.update({ key: value }))) for key in kwargs]
        self._freeze()

    @staticmethod
    def set_default(
            morph_region_method: BCsVolumetricModality = None,
            morphable_layers: int = None,
            morph_region_box_extension: float = None):
        """Set the default values of MorphBCParams.

        Parameters
        ----------
        morph_region_method: BCsVolumetricModality, optional
            Indicate options to identify morphable regions in the input volumetric mesh.
        morphable_layers: int, optional
            Number of layers around defined boundary.
        morph_region_box_extension: float, optional
            Percentage extension for the volumetric box.
        """
        args = locals()
        [MorphBCParams._default_params.update({ key: value }) for key, value in args.items() if value is not None]

    @staticmethod
    def print_default():
        """Print the default values of MorphBCParams.

        Examples
        --------
        >>> MorphBCParams.print_default()
        """
        message = ""
        message += ''.join(str(key) + ' : ' + str(value) + '\n' for key, value in MorphBCParams._default_params.items())
        print(message)

    def _jsonify(self) -> Dict[str, Any]:
        json_data = {}
        if self._morph_region_method is not None:
            json_data["morphRegionMethod"] = self._morph_region_method
        if self._morphable_layers is not None:
            json_data["morphableLayers"] = self._morphable_layers
        if self._morph_region_box_extension is not None:
            json_data["morphRegionBoxExtension"] = self._morph_region_box_extension
        [ json_data.update({ utils.to_camel_case(key) : value }) for key, value in self._custom_params.items()]
        return json_data

    def __str__(self) -> str:
        message = "morph_region_method :  %s\nmorphable_layers :  %s\nmorph_region_box_extension :  %s" % (self._morph_region_method, self._morphable_layers, self._morph_region_box_extension)
        message += ''.join('\n' + str(key) + ' : ' + str(value) for key, value in self._custom_params.items())
        return message

    @property
    def morph_region_method(self) -> BCsVolumetricModality:
        """Indicate options to identify morphable regions in the input volumetric mesh.
        """
        return self._morph_region_method

    @morph_region_method.setter
    def morph_region_method(self, value: BCsVolumetricModality):
        self._morph_region_method = value

    @property
    def morphable_layers(self) -> int:
        """Number of layers around defined boundary.
        """
        return self._morphable_layers

    @morphable_layers.setter
    def morphable_layers(self, value: int):
        self._morphable_layers = value

    @property
    def morph_region_box_extension(self) -> float:
        """Percentage extension for the volumetric box.
        """
        return self._morph_region_box_extension

    @morph_region_box_extension.setter
    def morph_region_box_extension(self, value: float):
        self._morph_region_box_extension = value



"""Module for meshing utility functions."""
import enum
import os
from typing import Iterable, List

import ansys.meshing.prime as prime

from .scope import SurfaceScope, VolumeScope
from .utils import check_name_pattern


class LabelToZoneMethod(enum.IntEnum):
    """Provides for creating zones from labels."""

    SIMPLE = 0
    """
    Simple method for creating zones from labels.

    Entities are queried using labels, and zones are created.
    """


class Mesh:
    """Provides methods for users who are new to meshing.

    This class also serves as a tutorial for commonly used meshing workflows.

    The ``Mesh`` class provides these functionalities:

    * Surface meshing with constant and variable sizing with triangle
      or quad dominant mesh
    * Volume meshing with prism, tetrahedral, and polyhedral elements
    * Surface wrapping
    * Helper method for creating zones from labels
    * Helper methods for reading and writing files

    """

    def __init__(self, model: prime.Model):
        """Initialize the mesher using a model.

        Parameters
        ----------
        model : prime.Model
            Model that the methods are to work on.
        """
        self._model = model
        self._logger = model.python_logger

    def read(
        self, file_name: str, append: bool = False, cad_reader_route: prime.CadReaderRoute = None
    ):
        """Read or import files of different formats based on file extensions.

        This method supports reading and importing these file formats:

        * Reading PyPrimeMesh's native PMDAT files
        * Importing various types of CAD files
        * Importing Fluent Meshing's MSH files
        * Importing Fluent's CAS files
        * Importing MAPDL's CDB files

        Parameters
        ----------
        file_name : str
            Path to the file to read or import.
        append : bool, optional
            Whether to append to the file. The default is ``False``, in which case
             the file is overwritten.
        cad_reader_route : prime.CadReaderRoute, optional
            Route of the CAD reader. The default is ``None``.

        """
        filename, fileext = os.path.splitext(file_name)
        if fileext == ".msh" or file_name[-7:] == ".msh.gz":
            prime.FileIO(self._model).import_fluent_meshing_meshes(
                [file_name], prime.ImportFluentMeshingMeshParams(self._model, append=append)
            )
        elif fileext == ".cas" or file_name[-7:] == ".cas.gz" or file_name[-7:] == ".cas.h5":
            prime.FileIO(self._model).import_fluent_case(
                file_name, prime.ImportFluentCaseParams(self._model, append=append)
            )
        elif fileext == ".cdb":
            prime.FileIO(self._model).import_mapdl_cdb(
                file_name, prime.ImportMapdlCdbParams(self._model, append=append)
            )
        elif fileext == ".pmdat":
            prime.FileIO(self._model).read_pmdat(
                file_name, prime.FileReadParams(self._model, append=append)
            )
        else:
            prime.FileIO(self._model).import_cad(
                file_name,
                prime.ImportCadParams(
                    self._model, append=append, cad_reader_route=cad_reader_route
                ),
            )

    def write(self, file_name: str):
        """Write or export files of different formats based on file extensions.

        This method supports writing and exporting these file formats:

        * Writing PyPrimeMesh's native PMDAT files
        * Exporting Fluent Meshing's MSH files
        * Exporting Fluent's CAS files
        * Exporting MAPDL's CDB files.
        * Exporting STL files.

        Parameters
        ----------
        file_name : str
            Path of the file to write or export.

        """
        filename, fileext = os.path.splitext(file_name)
        if fileext == ".cdb":
            prime.FileIO(self._model).export_mapdl_cdb(
                file_name, prime.ExportMapdlCdbParams(self._model)
            )
        elif fileext == ".cas" or file_name[-7:] == ".cas.gz":
            prime.FileIO(self._model).export_fluent_case(
                file_name, prime.ExportFluentCaseParams(self._model, cff_format=False)
            )
        elif file_name[-7:] == ".cas.h5":
            prime.FileIO(self._model).export_fluent_case(
                file_name, prime.ExportFluentCaseParams(self._model, cff_format=True)
            )
        elif fileext == ".msh" or file_name[-7:] == ".msh.gz":
            prime.FileIO(self._model).export_fluent_meshing_mesh(
                file_name, prime.ExportFluentMeshingMeshParams(self._model)
            )
        elif fileext == ".stl":
            part_ids = [part.id for part in self._model.parts]
            prime.FileIO(self._model).export_stl(
                file_name, prime.ExportSTLParams(self._model, part_ids)
            )
        else:
            prime.FileIO(self._model).write_pmdat(file_name, prime.FileWriteParams(self._model))

    def create_zones_from_labels(
        self,
        label_expression: str = None,
        conversion_method: LabelToZoneMethod = LabelToZoneMethod.SIMPLE,
    ):
        """Create zones from labels.

        When exporting to various solvers, zones play a very important role.
        Zones are where material properties and boundary conditions
        can be set in respective solvers.

        Zones allow downstream setting of boundary conditions or material properties.

        Zone names in PyPrimeMesh are translated into equivalent concepts in the solver.
        Currently, only one method is available to convert a zone to a label, and
        only face zones are created.

        The ``create_zones_from_labels`` method finds the entities by labels
        and then adds them to zones with the same names as the labels.

        If the ``label_expression`` parameter is not provided, all labels are
        flattened to create zones. If overlaps occur, label names are combined, and
        separate zones are created.

        For example, if "LabelA" and "LabelB" have overlapping TopoFaces,
        these zones are created: ``"LabelA"``, ``"LabelB"``, and
        ``"LabelA_LabelB"`` for the overlap.

        Parameters
        ----------
        label_expression : str, optional
            Expression for the labels to convert to zones. The
            default is ``None``.
        conversion_method : LabelToZoneMethod, optional
            Method for converting the labels to zones. The default
            is ``LabelToZoneMethod.SIMPLE``.

        Examples
        --------
        >>> from ansys.meshing.prime import lucid
        >>> mesh_util = lucid.Mesh(model)
        >>> mesh_util.create_zones_from_labels()

        """
        if conversion_method != LabelToZoneMethod.SIMPLE:
            self._logger.error("Invalid label to zone conversion method")

        if label_expression:
            labels = []
            for part in self._model.parts:
                lbls = part.get_labels()
                for l in lbls:
                    if l not in labels and check_name_pattern(label_expression, l):
                        labels.append(l)

            for label in labels:
                face_zone = None
                for part in self._model.parts:
                    topo_faces = part.get_topo_faces()
                    if len(topo_faces) > 0:
                        in_topo_faces = part.get_topo_faces_of_label_name_pattern(
                            label, prime.NamePatternParams(model=self._model)
                        )
                        if len(in_topo_faces) > 0:
                            if face_zone == None:
                                face_zone = self._model.create_zone(label, prime.ZoneType.FACE)
                            part.add_topo_entities_to_zone(face_zone.zone_id, in_topo_faces)
                    else:
                        in_face_zonelets = part.get_face_zonelets_of_label_name_pattern(
                            label, prime.NamePatternParams(model=self._model)
                        )
                        if len(in_face_zonelets) > 0:
                            if face_zone == None:
                                face_zone = self._model.create_zone(label, prime.ZoneType.FACE)
                            part.add_zonelets_to_zone(face_zone.zone_id, in_face_zonelets)
        else:
            labels = []
            for part in self._model.parts:
                labels += part.get_labels()
            faces_of_label = {}
            labels_of_face = {}
            label_zone_definitions = {}
            lists_of_labels = []
            face_zones = []
            all_topo = []
            for part in self._model.parts:
                for face in part.get_topo_faces():
                    labels_of_face[face] = []
                for face in part.get_face_zonelets():
                    labels_of_face[face] = []
            for label in labels:
                faces = []
                faces_of_label[label] = []
                for part in self._model.parts:
                    for zone in part.get_face_zones():
                        face_zones.append(self._model.get_zone_name(zone))
                    topo_faces = part.get_topo_faces()
                    all_topo += topo_faces
                    if topo_faces:
                        faces += part.get_topo_faces_of_label_name_pattern(
                            label_name_pattern=label,
                            name_pattern_params=prime.NamePatternParams(self._model),
                        )
                    else:
                        faces += part.get_face_zonelets_of_label_name_pattern(
                            label_name_pattern=label,
                            name_pattern_params=prime.NamePatternParams(self._model),
                        )
                if faces:
                    faces_of_label[label] += faces
                    [labels_of_face[face].append(label) for face in faces]
            for face in labels_of_face:
                name_exists = False
                for label_list in lists_of_labels:
                    zone_name = '_'.join(labels_of_face[face])
                    if set(label_list) == set(labels_of_face[face]):
                        name_exists = True
                        label_zone_definitions[zone_name].append(face)
                if not name_exists:
                    lists_of_labels.append(labels_of_face[face])
                    label_zone_definitions['_'.join(labels_of_face[face])] = [face]
            # remove empty labels
            if "" in label_zone_definitions:
                label_zone_definitions.pop("")
            self._logger.info("Labels to zones: " + str(label_zone_definitions))
            for zone_name in label_zone_definitions:
                if zone_name not in face_zones:
                    self._model.create_zone(zone_name, prime.ZoneType.FACE)
                for face in label_zone_definitions[zone_name]:
                    for part in self._model.parts:
                        if face in all_topo and face in part.get_topo_faces():
                            part.add_topo_entities_to_zone(
                                topo_entities=[face],
                                zone_id=self._model.get_zone_by_name(zone_name=zone_name),
                            )
                        elif face not in all_topo and face in part.get_face_zonelets():
                            part.add_zonelets_to_zone(
                                zonelets=[face],
                                zone_id=self._model.get_zone_by_name(zone_name=zone_name),
                            )

    def merge_parts(self, parts_expression: str = "*", new_name: str = "merged_part"):
        """Merge multiple parts into a single part.

        Parameters
        ----------
        parts_expression : str, optional
            Expression of the parts to merge. The default is ``"*"``.
        new_name : str, optional
            Name of the new part that is created from the merged parts.
            The default is ``"merged_part"``.

        """
        part_ids = []
        for part in self._model.parts:
            if check_name_pattern(parts_expression, part.get_name()):
                part_ids.append(part.id)

        self._model.merge_parts(
            part_ids=part_ids,
            params=prime.MergePartsParams(self._model, merged_part_suggested_name=new_name),
        )

    def __surface_mesh_on_active_sf(self, generate_quads: bool, scope: SurfaceScope):
        part_ids = scope.get_parts(self._model)
        for part_id in part_ids:
            topofaces = scope.get_topo_faces(self._model, part_id)
            surfer = prime.Surfer(self._model)
            params = prime.SurferParams(
                self._model,
                size_field_type=prime.SizeFieldType.VOLUMETRIC,
                generate_quads=generate_quads,
            )
            if len(topofaces) > 0:
                surfer.mesh_topo_faces(part_id=part_id, topo_faces=topofaces, params=params)
            else:
                tf = scope.get_face_zonelets(self._model, part_id)
                if len(tf) > 0:
                    te = []
                    surfer.remesh_face_zonelets(part_id, tf, te, params)

    def __variable_size_surface_mesh(
        self, min_size: float, max_size: float, generate_quads: bool, scope: SurfaceScope
    ):
        self._model.set_global_sizing_params(
            prime.GlobalSizingParams(model=self._model, min=min_size, max=max_size, growth_rate=1.2)
        )
        surfer = prime.Surfer(self._model)
        params = prime.SurferParams(
            self._model,
            size_field_type=prime.SizeFieldType.VOLUMETRIC,
            generate_quads=generate_quads,
        )
        part_ids = scope.get_parts(self._model)
        for part_id in part_ids:
            part = self._model.get_part(part_id)
            topofaces = scope.get_topo_faces(self._model, part_id)
            if len(topofaces) > 0:
                sizecontrol1 = self._model.control_data.create_size_control(
                    prime.SizingType.CURVATURE
                )
                sizecontrol1.set_curvature_sizing_params(
                    prime.CurvatureSizingParams(
                        model=self._model, min=min_size, max=max_size, growth_rate=1.2
                    )
                )
                sizecontrol1.set_scope(
                    prime.ScopeDefinition(
                        model=self._model,
                        entity_type=prime.ScopeEntity.FACEANDEDGEZONELETS,
                        part_expression=part.name,
                    )
                )
                sizefield = prime.SizeField(model=self._model)
                res = sizefield.compute_volumetric(
                    size_control_ids=[sizecontrol1.id],
                    volumetric_sizefield_params=prime.VolumetricSizeFieldComputeParams(
                        model=self._model, enable_multi_threading=False
                    ),
                )
                surfer.mesh_topo_faces(part_id=part.id, topo_faces=topofaces, params=params)
                self._model.delete_volumetric_size_fields([res.size_field_id])
            else:
                tf = scope.get_face_zonelets(self._model, part_id)
                if len(tf) > 0:
                    te = []
                    sizecontrol1 = self._model.control_data.create_size_control(
                        prime.SizingType.CURVATURE
                    )
                    sizecontrol1.set_curvature_sizing_params(
                        prime.CurvatureSizingParams(
                            model=self._model, min=min_size, max=max_size, growth_rate=1.2
                        )
                    )
                    sizecontrol1.set_scope(
                        prime.ScopeDefinition(
                            model=self._model,
                            entity_type=prime.ScopeEntity.FACEANDEDGEZONELETS,
                            part_expression=part.name,
                        )
                    )
                    sizefield = prime.SizeField(model=self._model)
                    res = sizefield.compute_volumetric(
                        size_control_ids=[sizecontrol1.id],
                        volumetric_sizefield_params=prime.VolumetricSizeFieldComputeParams(
                            model=self._model, enable_multi_threading=False
                        ),
                    )
                    surfer.remesh_face_zonelets(part.id, tf, te, params)
                    self._model.delete_volumetric_size_fields([res.size_field_id])

    def __constant_size_surface_mesh(
        self, min_size: float, generate_quads: bool, scope: SurfaceScope
    ):
        part_ids = scope.get_parts(self._model)
        surfer = prime.Surfer(self._model)
        for part_id in part_ids:
            topofaces = scope.get_topo_faces(self._model, part_id)
            params = prime.SurferParams(
                self._model, constant_size=min_size, generate_quads=generate_quads
            )
            if len(topofaces) > 0:
                surfer.mesh_topo_faces(part_id=part_id, topo_faces=topofaces, params=params)
            else:
                tf = scope.get_face_zonelets(self._model, part_id)
                te = []
                surfer.remesh_face_zonelets(part_id, tf, te, params)

    def surface_mesh(
        self,
        min_size: float = None,
        max_size: float = None,
        generate_quads: bool = False,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate a surface mesh on a scope.

        This method is used to generate a surface mesh of a constant or variable
        size. It supports generating quad dominant or triangular elements.

        If minimum and maximum sizes are provided, a variable size mesh is generated
        between the minimum size and maximum size by applying sizes based on curvature.

        If only the minimum size or maximum size is provided, a constant size mesh
        is generated with the provided size.

        If neither minimum size or maximum size is provided, the global maximum setting
        is used to generate a constant size mesh.

        Parameters
        ----------
        min_size : float, optional
            Minimum edge length of the mesh. The default is ``None``.
        max_size : float, optional
            Maximum edge length of the mesh. The default is ``None``.
        generate_quads : bool, optional
            Whether to generate a quad dominant mesh. The default is
            ``False``, in which case a triangular mesh is generated.
        scope : SurfaceScope
            Scope for generating the surface mesh.

        """
        if min_size == None and max_size == None:
            global_sizing = self._model.get_global_sizing_params()
            self._logger.warning(
                "Min and Max size not provided. Using max global size "
                + str(global_sizing.max)
                + " and min global size "
                + str(global_sizing.min)
            )
            self.__variable_size_surface_mesh(
                min_size=global_sizing.min,
                max_size=global_sizing.max,
                generate_quads=generate_quads,
                scope=scope,
            )
        elif min_size == None or max_size == None:
            if min_size is None:
                self.__constant_size_surface_mesh(
                    min_size=max_size, generate_quads=generate_quads, scope=scope
                )
            else:
                self.__constant_size_surface_mesh(
                    min_size=min_size, generate_quads=generate_quads, scope=scope
                )
        else:
            self.__variable_size_surface_mesh(
                min_size=min_size, max_size=max_size, generate_quads=generate_quads, scope=scope
            )

    def create_constant_size_control(
        self,
        control_name: str = "size_control",
        size: float = 1.0,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate constant size control on a scope.

        Parameters
        ----------
        control_name : str
            Name of the control.
        size : float, optional
            Constant edge length of the mesh. The default is ``1.0``.
        scope : SurfaceScope
            Scope for creating the size control.

        """
        global_sizes = self._model.get_global_sizing_params()
        if global_sizes.max < size:
            global_sizes.max = size
        if global_sizes.min > size or global_sizes.min <= 0:
            global_sizes.min = size
        self._model.set_global_sizing_params(global_sizes)
        sizecontrol = self._model.control_data.create_size_control(prime.SizingType.SOFT)
        sizecontrol.set_soft_sizing_params(
            prime.SoftSizingParams(
                model=self._model, max=size, growth_rate=global_sizes.growth_rate
            )
        )
        sizecontrol.set_scope(scope.get_scope_definition(self._model))
        sizecontrol.set_suggested_name(control_name)

    def create_curvature_size_control(
        self,
        control_name: str = "size_control",
        min: float = 1.0,
        max: float = 2.0,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate a curvature size control on a scope.

        Parameters
        ----------
        control_name : str, optional
            Name of the control. The default is ``"size_control"``.
        min : float, optional
            Minimum edge length of the mesh. The default is ``1.0``.
        max : float
            Maximum edge length of the mesh. The default is ``2.0``.
        scope : SurfaceScope
            Scope for creating the size control.

        """
        global_sizes = self._model.get_global_sizing_params()
        if max < min:
            t = max
            max = min
            min = t
        if global_sizes.max < max:
            global_sizes.max = max
        if global_sizes.min > min or global_sizes.min <= 0:
            global_sizes.min = min
        self._model.set_global_sizing_params(global_sizes)
        sizecontrol = self._model.control_data.create_size_control(prime.SizingType.CURVATURE)
        sizecontrol.set_curvature_sizing_params(
            prime.CurvatureSizingParams(
                model=self._model, min=min, max=max, growth_rate=global_sizes.growth_rate
            )
        )
        sizecontrol.set_scope(scope.get_scope_definition(self._model))
        sizecontrol.set_suggested_name(control_name)

    def surface_mesh_with_size_controls(
        self,
        size_control_names: str = "*",
        generate_quads: bool = False,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate a surface mesh on a scope using the given size controls.

        This method supports generating quad dominant or triangular elements.

        Parameters
        ----------
        size_control_names : str, optional
            Name pattern for the size controls. The default is ``"*"``.
        generate_quads : bool, optional
            Whether to generate a quad dominant mesh. The default is
            ``False``, in which case a triangular mesh is generated.
        scope : SurfaceScope
            Scope for generating the surface mesh.

        """
        sizefield = prime.SizeField(model=self._model)
        s_control_ids = []
        s_controls = self._model.control_data.size_controls
        for control in s_controls:
            if check_name_pattern(size_control_names, control.name):
                s_control_ids.append(control.id)
        if len(s_control_ids) > 0:
            size_field_res = sizefield.compute_volumetric(
                size_control_ids=s_control_ids,
                volumetric_sizefield_params=prime.VolumetricSizeFieldComputeParams(
                    model=self._model, enable_multi_threading=False
                ),
            )
            self.__surface_mesh_on_active_sf(generate_quads=generate_quads, scope=scope)
            self._model.delete_volumetric_size_fields([size_field_res.size_field_id])

    def connect_faces(
        self,
        part_expression: str = "*",
        face_labels: str = "*",
        target_face_labels: str = "*",
        tolerance: float = 0.05,
    ):
        """Connect face zonelets with a label name pattern within a given tolerance.

        This method connects face zonelets with a label name pattern to
        face zonelets with target face labels within a given tolerance. The connection
        happens within the part. Face zonelets of a part are connected with face zonelets
        of the same part only.

        Parameters
        ----------
        part_expression : str, optional
            Name pattern of the parts to use for connecting the face zonelets. The
            default is ``"*"``.
        face_labels : str, optional
            Name pattern of the face labels used for connecting the face zonelets.
            The default is ``"*"``.
        target_face_labels : str, optional
            Name pattern of the face labels to connect. The default is ``"*"``.
        tolerance: float, optional
            Tolerance for the connection. The default is ``0.05``.

        """
        name_pattern_param = prime.NamePatternParams(self._model)
        connect = prime.Connect(self._model)
        for part in self._model.parts:
            if check_name_pattern(part_expression, part.name):
                join_faces_source = part.get_face_zonelets_of_label_name_pattern(
                    face_labels, name_pattern_param
                )
                join_faces_target = part.get_face_zonelets_of_label_name_pattern(
                    target_face_labels, name_pattern_param
                )
                if len(join_faces_source) > 0 and len(join_faces_target) > 0:
                    connect.intersect_face_zonelets(
                        part.id,
                        face_zonelet_ids=join_faces_source,
                        with_face_zonelet_ids=join_faces_target,
                        params=prime.IntersectParams(self._model, tolerance=tolerance),
                    )
                    connect.join_face_zonelets(
                        part.id,
                        face_zonelet_ids=join_faces_source,
                        with_face_zonelet_ids=join_faces_target,
                        params=prime.JoinParams(self._model, tolerance=tolerance),
                    )

    def compute_volumes(self, part_expression: str = "*", create_zones_per_volume: bool = True):
        """Compute volumes in the parts defined by the part expression.

        Parameters
        ----------
        part_expression : str, optional
            Expression of the parts where the topology must be deleted.
            The default is ``"*"``.
        create_zones_per_volume : bool, optional
            Whether to create volume zones for each volume. The default
            is ``True``.

        """
        create_zones_type = prime.CreateVolumeZonesType.NONE
        if create_zones_per_volume:
            create_zones_type = prime.CreateVolumeZonesType.PERVOLUME
        params = prime.ComputeVolumesParams(model=self._model, create_zones_type=create_zones_type)
        for part in self._model.parts:
            if check_name_pattern(part_expression, part.name):
                if len(part.get_topo_faces()) > 0:
                    part.compute_topo_volumes(params=params)
                else:
                    part.compute_closed_volumes(params=params)

    def delete_topology(self, part_expression: str = "*", delete_edges: bool = True):
        """Delete topology in one or more parts.

        Parameters
        ----------
        part_expression : str, optional
            Expression of the parts where the topology must be deleted.
            The default is ``"*"``.
        delete_edges : bool, optional
            Whether to delete edge zonelets. The default is ``True``.

        """
        for part in self._model.parts:
            if check_name_pattern(part_expression, part.name):
                part.delete_topo_entities(
                    prime.DeleteTopoEntitiesParams(
                        model=self._model, delete_geom_zonelets=True, delete_mesh_zonelets=False
                    )
                )
                if delete_edges:
                    part.delete_zonelets(part.get_edge_zonelets())

    def __create_cap(self, part_id: int, face_zonelets: Iterable[int]):
        su = prime.SurfaceUtilities(model=self._model)
        res = su.create_cap_on_face_zonelets(
            part_id=part_id,
            face_zonelets=face_zonelets,
            params=prime.CreateCapParams(model=self._model),
        )
        return res.created_face_zonelets

    def __compute_flow_volume(
        self, part_id: int, face_zonelets: Iterable[int], flow_volume_zone_name: str = "flow_volume"
    ):
        part = self._model.get_part(part_id)
        if part:
            params = prime.ExtractVolumesParams(
                model=self._model, create_zone=True, suggested_zone_name=flow_volume_zone_name
            )
            res = part.extract_volumes(face_zonelets=face_zonelets, params=params)

    def create_flow_volume(
        self, flow_volume_zone_name: str = "flow_volume", cap_scope: SurfaceScope = SurfaceScope()
    ):
        """Create flow volume by the face labels defining the boundary of the volume.

        Parameters
        ----------
        flow_volume_zone_name : str, optional
            Name for the volume zone of the created flow volume. The default is
            ``"flow_volume"``.
        cap_scope : SurfaceScope
            Scope defining the face zonelets where the cap for flow volume must
            be created.

        """
        parts = cap_scope.get_parts(self._model)
        for part in parts:
            caps = self.__create_cap(part, cap_scope.get_face_zonelets(self._model, part))
            if len(caps) > 0:
                self.__compute_flow_volume(part, caps, flow_volume_zone_name)

    def __create_volume_controls(self, part: prime.Part, scope: VolumeScope) -> Iterable[int]:
        volume_control_ids = []
        if scope._evaluation_type == prime.ScopeEvaluationType.ZONES:
            volume_zones_all = part.get_volume_zones()
            volume_zones_to_mesh = []
            for volume_zone in volume_zones_all:
                if check_name_pattern(
                    scope._entity_expression, self._model.get_zone_name(volume_zone)
                ):
                    volume_zones_to_mesh.append(volume_zone)

            volume_zones_to_avoid = [v for v in volume_zones_all if v not in volume_zones_to_mesh]
            scope_str = ",".join([self._model.get_zone_name(v) for v in volume_zones_to_avoid])

            if len(scope_str) > 0:
                volume_control_param_dead = prime.VolumeControlParams(
                    model=self._model, cell_zonelet_type=prime.CellZoneletType.DEAD
                )
                volume_control_scope_dead = prime.ScopeDefinition(
                    model=self._model,
                    part_expression=part.name,
                    label_expression="",
                    zone_expression=scope_str,
                    entity_type=prime.ScopeEntity.VOLUME,
                    evaluation_type=prime.ScopeEvaluationType.ZONES,
                )
                volume_control_dead = self._model.control_data.create_volume_control()
                volume_control_dead.set_params(volume_control_params=volume_control_param_dead)
                volume_control_dead.set_scope(scope=volume_control_scope_dead)
                volume_control_ids.append(volume_control_dead.id)

            volume_control_param_fluid = prime.VolumeControlParams(
                model=self._model, cell_zonelet_type=prime.CellZoneletType.FLUID
            )
            volume_control_scope_fluid = prime.ScopeDefinition(
                model=self._model,
                part_expression=part.name,
                label_expression="",
                zone_expression=scope._entity_expression,
                entity_type=prime.ScopeEntity.VOLUME,
                evaluation_type=prime.ScopeEvaluationType.ZONES,
            )
            volume_control_fluid = self._model.control_data.create_volume_control()
            volume_control_fluid.set_params(volume_control_params=volume_control_param_fluid)
            volume_control_fluid.set_scope(scope=volume_control_scope_fluid)
            volume_control_ids.append(volume_control_fluid.id)

        return volume_control_ids

    def volume_mesh(
        self,
        volume_fill_type: prime.VolumeFillType = prime.VolumeFillType.TET,
        quadratic: bool = False,
        prism_layers: int = None,
        prism_surface_expression: str = "*",
        prism_volume_expression: str = "*",
        growth_rate: float = 1.2,
        scope: VolumeScope = VolumeScope(),
    ):
        """Generate a volume mesh on the model.

        Parameters
        ----------
        volume_fill_type : prime.VolumeFillType
            Type of volume elements to generate.
        quadratic : bool, optional
            Whether to generate a quadratic mesh. The default is ``False``.
            A quadratic mesh is not supported with parallel meshing.
            It is only supported with a pure tetrahedral mesh.
        prism_layers : int, optional
            Number of prism layers to grow. The default is ``None``.
            If a value for prism layers is specified, prism layers are generated.
        prism_surface_expression : str, optional
            Face zones or TopoFaces to grow prisms from. The default is ``"*"``,
            in which case prisms are grown from all surfaces.
        prism_volume_expression : str, optional
            Volumes or TopoVolumes to grow prisms in.
            The default is ``"*"``. This expression evaluates
            to zone names and volumes, or TopoVolumes are queried
            based on the zones evaluated.
        growth_rate : float, optional
            Prism growth rate. The default is ``1.2``.
        scope : VolumeScope
            Scope of volumes to mesh.

        """
        automesh_params = prime.AutoMeshParams(model=self._model)
        automesh_params.volume_fill_type = volume_fill_type
        if quadratic:
            automesh_params.tet = prime.TetParams(self._model, True)
        automesh = prime.AutoMesh(self._model)
        for part in self._model.parts:
            if check_name_pattern(scope._part_expression, part.name):
                try:
                    prism_control = None
                    volume_control_ids = self.__create_volume_controls(part, scope)
                    if len(volume_control_ids) > 0:
                        automesh_params.volume_control_ids = volume_control_ids
                    if prism_layers:
                        prism_control = self._model.control_data.create_prism_control()
                        prism_control.set_surface_scope(
                            prime.ScopeDefinition(
                                self._model,
                                part_expression=part.name,
                                label_expression=prism_surface_expression,
                            )
                        )
                        prism_control.set_growth_params(
                            prime.PrismControlGrowthParams(
                                self._model, n_layers=prism_layers, growth_rate=growth_rate
                            )
                        )
                        prism_control.set_volume_scope(
                            prime.ScopeDefinition(
                                self._model,
                                entity_type=prime.ScopeEntity.VOLUME,
                                evaluation_type=prime.ScopeEvaluationType.ZONES,
                                part_expression=part.name,
                                zone_expression=prism_volume_expression,
                            )
                        )
                        automesh_params.prism_control_ids = [prism_control.id]

                    automesh.mesh(part.id, automesh_params=automesh_params)
                    if prism_control:
                        self._model.control_data.delete_controls([prism_control.id])
                    if len(volume_control_ids) > 0:
                        self._model.control_data.delete_controls(volume_control_ids)
                except:
                    self._logger.info(part.name + " not volume meshed.")

    def __setup_sizing_for_wrapper(
        self,
        min_size: float,
        max_size: float,
        growth_rate: float,
        normal_angle: float,
        elements_per_gap: float,
        global_size_controls: List,
        geodesic_global_size_controls: List,
        scope: prime.ScopeDefinition,
        global_sf: prime.GlobalSizingParams,
        wrap_size_controls: List[prime.SizeControl],
    ):
        constant_size = None
        if min_size and not max_size:
            max_size = min_size
            constant_size = min_size
        elif max_size and not min_size:
            min_size = max_size
            constant_size = max_size

        if not max_size and not min_size:
            max_size = global_sf.max
            min_size = global_sf.min
        if not growth_rate:
            growth_rate = global_sf.growth_rate
        self._model.set_global_sizing_params(
            prime.GlobalSizingParams(
                self._model, min=min_size, max=max_size, growth_rate=growth_rate
            )
        )
        self._logger.info(
            "Min size: "
            + str(min_size)
            + ", Max size: "
            + str(max_size)
            + ", Growth rate: "
            + str(growth_rate)
        )

        is_geodesic = self.__is_size_field_type_geodesic(wrap_size_controls)
        if not constant_size:
            if normal_angle:
                curv_size_control = self._model.control_data.create_size_control(
                    sizing_type=prime.SizingType.CURVATURE
                )
                curv_size_control.set_curvature_sizing_params(
                    prime.CurvatureSizingParams(
                        model=self._model,
                        max=max_size,
                        min=min_size,
                        growth_rate=growth_rate,
                        normal_angle=normal_angle,
                    )
                )
                curv_size_control.set_scope(scope)
                global_size_controls.append(curv_size_control)
                if is_geodesic or elements_per_gap:
                    curv_size_control = self._model.control_data.create_size_control(
                        sizing_type=prime.SizingType.CURVATURE
                    )
                    curv_size_control.set_curvature_sizing_params(
                        prime.CurvatureSizingParams(
                            model=self._model,
                            max=max_size,
                            min=min_size,
                            growth_rate=growth_rate,
                            normal_angle=normal_angle,
                        )
                    )
                    curv_size_control.set_scope(scope)
                    geodesic_global_size_controls.append(curv_size_control)
            else:
                if elements_per_gap == None and not wrap_size_controls:
                    prime.PrimeRuntimeError(
                        "Error: No size functions or settings provided.\
                        No wrap done."
                    )
                    return
            if elements_per_gap:
                prox_size_control = self._model.control_data.create_size_control(
                    sizing_type=prime.SizingType.PROXIMITY
                )
                prox_size_control.set_proximity_sizing_params(
                    prime.ProximitySizingParams(
                        model=self._model,
                        elements_per_gap=elements_per_gap,
                        max=max_size,
                        min=min_size,
                        growth_rate=growth_rate,
                        ignore_orientation=True,
                        ignore_self_proximity=False,
                    )
                )
                prox_size_control.set_scope(scope)
                global_size_controls.append(prox_size_control)
        else:
            const_size_control = self._model.control_data.create_size_control(
                sizing_type=prime.SizingType.CURVATURE
            )
            const_size_control.set_curvature_sizing_params(
                prime.CurvatureSizingParams(
                    model=self._model,
                    max=constant_size,
                    min=constant_size,
                    growth_rate=growth_rate,
                    normal_angle=normal_angle,
                )
            )
            const_size_control.set_scope(scope)
            global_size_controls.append(const_size_control)
            if is_geodesic or elements_per_gap:
                const_size_control = self._model.control_data.create_size_control(
                    sizing_type=prime.SizingType.CURVATURE
                )
                const_size_control.set_curvature_sizing_params(
                    prime.CurvatureSizingParams(
                        model=self._model,
                        max=constant_size,
                        min=constant_size,
                        growth_rate=growth_rate,
                        normal_angle=normal_angle,
                    )
                )
                const_size_control.set_scope(scope)
                geodesic_global_size_controls.append(const_size_control)

        return min_size, max_size

    def __is_size_field_type_geodesic(self, controls: List[prime.SizeControl]):
        """Check if the size field type is geodesic."""
        geodesic = True
        geodesic_types = ["Curvature", "Soft"]
        self._logger.info("Size controls: " + str([control.name for control in controls]))
        for control in controls:
            result = control.get_summary(prime.SizeControlSummaryParams(self._model))
            message = str(result.message)
            ctrl_type = message.split("Type : ")[1]
            ctrl_type = ctrl_type.split("\n")[0]
            self._logger.info("Name: '" + control.name + "' Type: '" + ctrl_type + "'")
            if ctrl_type not in geodesic_types:
                geodesic = False
                self._logger.info("Note: VOLUMETRIC method used.")
                return geodesic
        return geodesic

    def __compute_size_field(
        self, size_controls: List[prime.SizeControl], field: prime.SizeField
    ) -> int:
        """Compute size field using the helper method."""
        self._model.delete_volumetric_size_fields(self._model.get_volumetric_size_fields())
        result = field.compute_volumetric(
            [size_control.id for size_control in size_controls],
            prime.VolumetricSizeFieldComputeParams(model=self._model, enable_multi_threading=False),
        )
        return result.size_field_id

    def __remesh_after_wrap(self, part: prime.Part):
        """Surface mesh after wrapping."""
        surfer = prime.Surfer(self._model)
        feature_edges = part.get_edge_zonelets_of_label_name_pattern(
            label_name_pattern="___wrapper_feature_path___",
            name_pattern_params=prime.NamePatternParams(model=self._model),
        )
        surf_params = surfer.initialize_surfer_params_for_wrapper()
        surf_params.size_field_type = prime.SizeFieldType.VOLUMETRIC
        surfer.remesh_face_zonelets(
            part_id=part.id,
            face_zonelets=part.get_face_zonelets(),
            edge_zonelets=feature_edges,
            params=surf_params,
        )

    def __create_contact_preventions(
        self,
        contact_prevention_params: List,
        contact_prevention_size: float,
        min_size: float,
        max_size: float,
        growth_rate: float,
        geodesic_global_size_controls: List,
        scope: str,
    ):
        """
        Create contact prevention controls for all parts.

        A proximity size control is created for remeshing.
        """
        all_parts_labels = len(self._model.parts)
        all_parts_labels += sum([len(part.get_labels()) for part in self._model.parts])
        if contact_prevention_size < min_size:
            prime.PrimeRuntimeError("Contact_prevention_size is less than min_size.")
            return
        self._logger.info(
            "Contact prevention enabled for:"
            + str([part.name for part in self._model.parts])
            + str([part.get_labels() for part in self._model.parts if part.get_labels()])
        )

        if all_parts_labels > 1:
            geo_prox_size_control = self._model.control_data.create_size_control(
                sizing_type=prime.SizingType.PROXIMITY
            )
            geo_prox_size_control.set_proximity_sizing_params(
                prime.ProximitySizingParams(
                    model=self._model,
                    elements_per_gap=3.0,
                    max=max_size,
                    min=min_size,
                    growth_rate=growth_rate,
                    ignore_orientation=False,
                )
            )
            geo_prox_size_control.set_scope(scope)
            geodesic_global_size_controls.append(geo_prox_size_control)
            for part in self._model.parts:
                source = prime.ScopeDefinition(model=self._model, part_expression=part.name)
                contact_prevention_params.append(
                    prime.ContactPreventionParams(
                        model=self._model, source_scope=source, size=contact_prevention_size
                    )
                )
                for label in part.get_labels():
                    if part.get_face_zonelets_of_label_name_pattern(
                        label_name_pattern=label,
                        name_pattern_params=prime.NamePatternParams(self._model),
                    ):
                        source = prime.ScopeDefinition(model=self._model, label_expression=label)
                        contact_prevention_params.append(
                            prime.ContactPreventionParams(
                                model=self._model, source_scope=source, size=contact_prevention_size
                            )
                        )
        else:
            self._logger.warning(
                "Contact prevention size specified but \
                insufficient parts and labels identified to define contact."
            )

    def __create_feature_recovery_params(
        self,
        feature_recovery_params: List[prime.FeatureRecoveryParams],
        extract_features: bool,
        create_intersection_loops: bool,
        use_existing_features: bool,
        feature_angle: float,
        enable_feature_octree_refinement: bool,
        scope: prime.ScopeDefinition,
    ):
        features = prime.FeatureExtraction(self._model)
        extracted_features = dict()
        if extract_features:
            face_zonelets_prime_array = self._model.control_data.get_part_zonelets(scope=scope)
            for item in face_zonelets_prime_array:
                res = features.extract_features_on_face_zonelets(
                    part_id=item.part_id,
                    face_zonelets=item.face_zonelets,
                    params=prime.ExtractFeatureParams(
                        model=self._model,
                        feature_angle=feature_angle,
                        separation_angle=20,
                        replace=True,
                        label_name="__extracted__features__",
                    ),
                )
                extracted_features[item.part_id] = res.new_edge_zonelets
            feature_scope = scope
            feature_scope.label_expression = "__extracted__features__"
            extracted_feature_params = prime.FeatureRecoveryParams(
                model=self._model,
                enable_feature_octree_refinement=enable_feature_octree_refinement,
                scope=feature_scope,
            )
            feature_recovery_params.append(extracted_feature_params)
        if create_intersection_loops:
            for part in self._model.parts:
                others = [item.name for item in self._model.parts if item.name != part.name]
                scope1 = prime.ScopeDefinition(self._model, part_expression=part.name)
                scope2 = prime.ScopeDefinition(self._model, part_expression=','.join(others))
                scope1_zonelets_array = self._model.control_data.get_part_zonelets(scope=scope1)
                scope2_zonelets_array = self._model.control_data.get_part_zonelets(scope=scope2)
                intersect_loops_params = prime.CreateIntersectionEdgeLoopsParams(
                    model=self._model, label_name="__intersect_loops__"
                )
                features.create_intersection_edge_loops(
                    part_face_zonelets=scope1_zonelets_array,
                    intersecting_part_face_zonelets=scope2_zonelets_array,
                    params=intersect_loops_params,
                )
            intersect_feature_params = prime.FeatureRecoveryParams(
                model=self._model,
                enable_feature_octree_refinement=enable_feature_octree_refinement,
                scope=prime.ScopeDefinition(
                    model=self._model, part_expression="*", label_expression="__intersect_loops__"
                ),
            )
            feature_recovery_params.append(intersect_feature_params)
        if use_existing_features:
            existing_feature_params = prime.FeatureRecoveryParams(
                model=self._model,
                enable_feature_octree_refinement=enable_feature_octree_refinement,
                scope=scope,
            )
            feature_recovery_params.append(existing_feature_params)
        return extracted_features

    def __process_size_fields(
        self,
        global_size_controls: List[prime.SizeControl],
        wrap_size_controls: List[prime.SizeControl],
        geodesic_global_size_controls: List[prime.SizeControl],
        remesh_postwrap: bool,
        use_existing_size_fields: bool,
        recompute_remesh_sizes: bool,
        wrap_params: prime.WrapParams,
        size_fields: List[int],
    ) -> List[int]:

        created_size_fields = []
        field = prime.SizeField(self._model)
        is_geodesic = self.__is_size_field_type_geodesic(global_size_controls + wrap_size_controls)
        if (
            is_geodesic
            and remesh_postwrap
            and not recompute_remesh_sizes
            and not use_existing_size_fields
        ):
            created_size_fields.append(
                self.__compute_size_field(geodesic_global_size_controls + wrap_size_controls, field)
            )

        if not use_existing_size_fields and (global_size_controls + wrap_size_controls):
            if is_geodesic:
                size_control_ids = [i.id for i in (global_size_controls + wrap_size_controls)]
                self._logger.info("Geodesic size controls used for wrap:" + str(size_control_ids))
                wrap_params.sizing_method = prime.SizeFieldType.GEODESIC
                wrap_params.size_control_ids = size_control_ids
            else:
                self._logger.info(
                    "Volumetric size controls used for wrap:"
                    + str([i.name for i in (global_size_controls + wrap_size_controls)])
                )
                field_id = self.__compute_size_field(
                    global_size_controls + wrap_size_controls, field
                )
                wrap_params.sizing_method = prime.SizeFieldType.VOLUMETRIC
                wrap_params.size_field_ids = [field_id]
                created_size_fields.append(field_id)
        else:
            if size_fields:
                self._logger.info(
                    "Using size fields provided for wrap:" + str([i for i in size_fields])
                )
                wrap_params.sizing_method = prime.SizeFieldType.VOLUMETRIC
                wrap_params.size_field_ids = size_fields
            else:
                size_field_ids = self._model.get_active_volumetric_size_fields()
                if size_field_ids:
                    self._logger.info("Existing size field used for wrap:" + str(size_field_ids))
                    wrap_params.sizing_method = prime.SizeFieldType.VOLUMETRIC
                    wrap_params.size_field_ids = size_field_ids
                else:
                    prime.PrimeRuntimeError("Invalid size fields")
        return created_size_fields

    def __process_after_wrap(
        self,
        remesh_postwrap: bool,
        wrapped_part: prime.Part,
        use_existing_size_fields: bool,
        global_size_controls: List[prime.SizeControl],
        remesh_size_controls: List[prime.SizeControl],
        geodesic_global_size_controls: List[prime.SizeControl],
        wrap_size_controls: List[prime.SizeControl],
        remesh_scope: str,
        recompute_remesh_sizes: bool,
        feature_angle: float,
        wrapper: prime.Wrapper,
    ) -> List[int]:

        created_size_fields = []
        field = prime.SizeField(self._model)
        is_geodesic = self.__is_size_field_type_geodesic(global_size_controls + wrap_size_controls)
        features = prime.FeatureExtraction(self._model)
        if remesh_postwrap and wrapped_part:
            if not use_existing_size_fields and (global_size_controls + remesh_size_controls):
                if recompute_remesh_sizes:

                    if is_geodesic:
                        self._logger.info(
                            "Recomputing sizing after geodesic wrap:"
                            + str(
                                [
                                    i.name
                                    for i in (geodesic_global_size_controls + remesh_size_controls)
                                ]
                            )
                        )
                        for i in range(len(geodesic_global_size_controls)):
                            control = geodesic_global_size_controls[i]
                            control.set_scope(remesh_scope)
                            geodesic_global_size_controls[i] = control
                        created_size_fields.append(
                            self.__compute_size_field(
                                geodesic_global_size_controls + remesh_size_controls, field
                            )
                        )
                    else:
                        self._logger.info(
                            "Recomputing sizing for remesh after volumetric wrap:"
                            + str([i.name for i in (global_size_controls + remesh_size_controls)])
                        )
                        for i in range(len(global_size_controls)):
                            control = global_size_controls[i]
                            control.set_scope(remesh_scope)
                            global_size_controls[i] = control
                        self.__compute_size_field(
                            global_size_controls + remesh_size_controls, field
                        )
                else:
                    if is_geodesic:
                        self._logger.info(
                            "Using geodesic size controls for remesh after geodesic wrap \
                            (consider using recompute_remesh_sizes):"
                            + str(
                                [
                                    i.name
                                    for i in (geodesic_global_size_controls + wrap_size_controls)
                                ]
                            )
                        )

            self.__remesh_after_wrap(wrapped_part)
            wrapper.improve_quality(
                part_id=wrapped_part.id,
                params=prime.WrapperImproveQualityParams(
                    model=self._model, island_count=20, target_skewness=0.9
                ),
            )
            return created_size_fields

    def wrap(
        self,
        min_size: float = None,
        max_size: float = None,
        growth_rate: float = 1.2,
        elements_per_gap: float = None,
        normal_angle: float = 18.0,
        input_parts: str = "*",
        input_labels: str = "*",
        keep_inputs: bool = False,
        region_extract: prime.WrapRegion = prime.WrapRegion.EXTERNAL,
        material_point: List[float] = None,
        extract_features: bool = True,
        create_intersection_loops: bool = False,
        use_existing_features: bool = False,
        enable_feature_octree_refinement: bool = True,
        feature_angle: float = 40.0,
        contact_prevention_size: float = None,
        number_of_threads: int = None,
        remesh_postwrap: bool = True,
        recompute_remesh_sizes: bool = False,
        use_existing_size_fields: bool = False,
        size_fields: List[prime.SizeField] = None,
        wrap_size_controls: List[prime.SizeControl] = None,
        remesh_size_controls: List[prime.SizeControl] = None,
        feature_recovery_params: List[prime.FeatureRecoveryParams] = None,
        contact_prevention_params: List[prime.ContactPreventionParams] = None,
        leak_prevention_params: List[prime.LeakPreventionParams] = None,
    ):
        """Wrap and remesh the input.

        The default behavior is to perform an external wrap of all parts in the model
        using curvature sizing and extracting features. The wrap is then remeshed
        to provide a surface mesh for the extracted region.

        Geodesic sizing is used only if soft and curvature controls are set.
        If contact prevention size is set and geodesic sizing is available, contact
        detection is used globally.

        If minimum and maximum sizes are provided, variable size mesh is generated
        between the minimum size and maximum size by applying sizes based on curvature.

        If either minimum size or maximum size is provided, constant size mesh
        is generated with the provided size.

        If neither minimum size or maximum size is provided, global minimum and
        maximum settings are used to generate a variable size mesh based on curvature.

        Parameters
        ----------
        min_size : float, optional
            Minimum edge length of the mesh. The default is ``None``.
        max_size : float, optional
            Maximum edge length of the mesh. The default is ``None``.
        growth_rate : float, optional
            Growth rate. The default is ``1.2``.
        elements_per_gap : float, optional
            Global proximity size control elements per gap with self proximity.
            The default is ``None``.
        normal_angle : float, optional
            Global curvature size control normal angle. The default is ``18.0``.
        input_parts : str, optional
            Parts to wrap. The default is ``"*"``.
        input_labels : str
            Labels to wrap. The default is ``"*"``.
        keep_inputs : bool, optional
            Whether to retain inputs. The default is ``False``.
        region_extract : prime.WrapRegion
            Region to wrap. The default is ``prime.WrapRegion.EXTERNAL``.
        material_point : List[float], optional
            Material point if the region extraction method is set to
            material point. The default is ``None``.
        extract_features : bool, optional
            Whether to extract feature edges using the feature angle. The
            default is ``True``.
        create_intersection_loops : bool, optional
            Whether to create intersection loops between all parts.
            The default is ``False``.
        use_existing_features : bool, optional
            Whether to maintain existing features on parts. The
            default is ``False``.
        enable_feature_octree_refinement: bool, optional
            Whether to apply refinement to feature edges during the wrap.
            The default is ``True``.
        feature_angle : float, optional
            Angle to use for extracting features. The default is ``40.0``.
        contact_prevention_size : float, optional
            Global proximity size controls between all parts.
            The default is ``None``.
        number_of_threads : int, optional
            Number of threads. The default is ``None``.
        remesh_postwrap : bool, optional
            Whether to remesh the wrap. The default is ``True``.
        recompute_remesh_sizes : bool, optional
            Whether to recompute sizes from global controls using
            the wrap surface. The default is ``False``.
        use_existing_size_fields : bool, optional
            Whether to use precomputed size fields. The default is ``False``.
        size_fields : List[prime.SizeField], optional
            List of size fields to use. The default is ``None``.
        wrap_size_controls : List[prime.SizeControl], optional
            List of wrap size controls to use. The default is ``None``.
        remesh_size_controls : List[prime.SizeControl], optional
            List of remesh size controls to use. The default is ``None``.
        feature_recovery_params : List[prime.FeatureRecoveryParams], optional
            List of feature recovery parameters to use. The default is ``None``.
        contact_prevention_params : List[prime.ContactPreventionParams], optional
            List of contact prevention parameters to use. The default is ``None``.
        leak_prevention_params : List[prime.LeakPreventionParams], optional
            List of leak prevention parameters to use.  The default is ``None``.

        Returns
        -------
        Wrapped part
                Returns Part.

        Examples
        --------
        >>> import ansys.meshing.prime as prime
        >>> from ansys.meshing.prime import lucid
        >>> prime_client = prime.launch_prime()
        >>> model = prime_client.model
        >>> mesh = lucid.Mesh(model)
        >>> mesh.read("/my_geometry.stl")
        >>> mesh.wrap(min_size=1, max_size=20, create_intersection_loops=True)
        >>> mesh.write("/mesh_output.pmdat")
        >>> prime_client.exit()

        """
        if size_fields is None:
            size_fields = []
        if wrap_size_controls is None:
            wrap_size_controls = []
        if remesh_size_controls is None:
            remesh_size_controls = []
        if feature_recovery_params is None:
            feature_recovery_params = []
        if contact_prevention_params is None:
            contact_prevention_params = []
        if leak_prevention_params is None:
            leak_prevention_params = []

        global_sf = prime.GlobalSizingParams(model=self._model)
        scope = prime.ScopeDefinition(
            model=self._model,
            part_expression=input_parts,
            label_expression=input_labels,
            entity_type=prime.ScopeEntity.FACEANDEDGEZONELETS,
        )

        # Delete topology and work only on zonelets for now
        for part in self._model.parts:
            if len(part.get_topo_faces()) > 0:
                part.delete_topo_entities(
                    prime.DeleteTopoEntitiesParams(
                        self._model, delete_geom_zonelets=False, delete_mesh_zonelets=True
                    )
                )

        params = prime.ScopeZoneletParams(model=self._model)
        zonelets = self._model.control_data.get_scope_face_zonelets(scope=scope, params=params)
        part_ids = self._model.control_data.get_scope_parts(scope=scope)

        global_size_controls = []
        geodesic_global_size_controls = []
        computed_size_fields = []

        min_size, max_size = self.__setup_sizing_for_wrapper(
            min_size=min_size,
            max_size=max_size,
            growth_rate=growth_rate,
            normal_angle=normal_angle,
            elements_per_gap=elements_per_gap,
            global_size_controls=global_size_controls,
            geodesic_global_size_controls=geodesic_global_size_controls,
            scope=scope,
            global_sf=global_sf,
            wrap_size_controls=wrap_size_controls,
        )

        wrapper_control = self._model.control_data.create_wrapper_control()
        wrap_params = prime.WrapParams(model=self._model, wrap_region=region_extract)
        if region_extract == prime.WrapRegion.MATERIALPOINT:
            if material_point:
                mp_result = self._model.material_point_data.create_material_point(
                    suggested_name="USER_MP",
                    coords=material_point,
                    params=prime.CreateMaterialPointParams(
                        model=self._model, type=prime.MaterialPointType.LIVE
                    ),
                )
                mp_name = mp_result.assigned_name
                wrapper_control.set_live_material_points([mp_name])
            else:
                prime.PrimeRuntimeError("Error: No material point provided. No wrap done.")
                return None

        wrapper_control.set_geometry_scope(scope)

        # Add contact preventions
        if contact_prevention_size:
            self.__create_contact_preventions(
                contact_prevention_params,
                contact_prevention_size,
                min_size,
                max_size,
                growth_rate,
                geodesic_global_size_controls,
                scope,
            )
        if contact_prevention_params:
            wrapper_control.set_contact_preventions(contact_prevention_params)

        # Add feature recoveries
        extracted_features = self.__create_feature_recovery_params(
            feature_recovery_params=feature_recovery_params,
            extract_features=extract_features,
            create_intersection_loops=create_intersection_loops,
            use_existing_features=use_existing_features,
            feature_angle=feature_angle,
            enable_feature_octree_refinement=enable_feature_octree_refinement,
            scope=scope,
        )
        if feature_recovery_params:
            wrapper_control.set_feature_recoveries(feature_recovery_params)

        # Add leak preventions
        if leak_prevention_params:
            wrapper_control.set_leak_preventions(leak_prevention_params)

        computed_size_fields += self.__process_size_fields(
            global_size_controls=global_size_controls,
            wrap_size_controls=wrap_size_controls,
            geodesic_global_size_controls=geodesic_global_size_controls,
            remesh_postwrap=remesh_postwrap,
            use_existing_size_fields=use_existing_size_fields,
            recompute_remesh_sizes=recompute_remesh_sizes,
            wrap_params=wrap_params,
            size_fields=size_fields,
        )

        # Wrap
        wrapper = prime.Wrapper(model=self._model)
        if number_of_threads:
            wrap_params.number_of_threads = number_of_threads
        res = wrapper.wrap(wrapper_control_id=wrapper_control.id, params=wrap_params)
        wrapped_part = self._model.get_part(res.id)
        remesh_scope = prime.ScopeDefinition(model=self._model, part_expression=wrapped_part.name)

        # Post Wrap Processing
        self.__process_after_wrap(
            remesh_postwrap,
            wrapped_part,
            use_existing_size_fields,
            global_size_controls,
            remesh_size_controls,
            geodesic_global_size_controls,
            wrap_size_controls,
            remesh_scope,
            recompute_remesh_sizes,
            feature_angle,
            wrapper,
        )

        volume_results = wrapped_part.compute_closed_volumes(
            prime.ComputeVolumesParams(
                model=self._model, create_zones_type=prime.CreateVolumeZonesType.PERVOLUME
            )
        )
        self._logger.info(str(volume_results.volumes) + " volumes found.")

        self._model.set_global_sizing_params(
            prime.GlobalSizingParams(
                self._model, min=global_sf.min, max=global_sf.max, growth_rate=global_sf.growth_rate
            )
        )
        self._model.control_data.delete_controls(
            [
                control.id
                for control in [wrapper_control]
                + global_size_controls
                + geodesic_global_size_controls
            ]
        )

        # delete extracted features
        for part_id, zonelets_to_delete in extracted_features.items():
            part = self._model.get_part(part_id)
            part.delete_zonelets(zonelets_to_delete)

        # delete size fields
        if len(computed_size_fields) > 0:
            self._model.delete_volumetric_size_fields(computed_size_fields)

        # retain or delete inputs
        if not keep_inputs:
            self._logger.info("Deleting inputs to wrap.")
            if zonelets:
                for part_id in part_ids:
                    part = self._model.get_part(part_id)
                    [
                        part.delete_zonelets([face])
                        for face in zonelets
                        if face in part.get_face_zonelets()
                    ]
            if part_ids:
                self._model.delete_parts(part_ids)

        self._logger.info("Wrap done.")
        return wrapped_part
